/**
	$Id: ChangeHistoryUtil.java 5073 2009-08-21 10:15:11Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import java.util.Date;

import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.ChangeHistory;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Session;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.User;
import net.sf.basedb.util.formatter.Formatter;

/**
	Utility class for working with change history.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2009-08-21 12:15:11 +0200 (Fri, 21 Aug 2009) $
*/
public class ChangeHistoryUtil
{

	private ChangeHistoryUtil()
	{}

	/**
		Check if the "Change history" tab should be visible for
		the current user or not. The tab is visible if change history
		logging has been enabled by the administrator and if the 
		current user has READ permission for CHANGEHISTORY.
		@param sc The current session control
		@return TRUE if the tab should be visible, FALSE otherwise
	*/
	public static final boolean showChangeHistoryTab(SessionControl sc)
	{
		return Config.getBoolean("changelog.show-in-web") && 
			sc.hasPermission(Permission.READ, Item.CHANGEHISTORY);
	}

	/**
		Get the item that was changed in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to the view page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getItem(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		BasicItem item = null;
		int itemId = ch.getItemId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		String result = null;
		try
		{
			item = ch.getItem(dc);
			String name = "";
			if (item instanceof File)
			{
				name = ((File)item).getPath().toString();
			}
			else if (item instanceof Nameable)
			{
				name = ((Nameable)item).getName();
			}
			else
			{
				name = item.toString();
			}
			if (link)
			{
				result = Base.getLink(ID, name, ch.getItemType(), itemId, editLink);
			}
			else
			{
				result = HTML.encodeTags(name);
			}
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		if (result == null)
		{
			result = getLink(ID, null, itemId, denied, notFound, false, false);
		}
		return result;
	}

	
	/**
		Get the user that made the change in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getUser(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		User item = null;
		int itemId = ch.getUserId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		try
		{
			item = ch.getUser(dc);
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		String result = getLink(ID, item, itemId, denied, notFound, link, editLink);
		return result;
	}
	
	/**
		Get the session in which the changes was made in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getSession(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		Session item = null;
		Integer itemId = ch.getSessionId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		String result = null;
		try
		{
			item = ch.getSession(dc);
			Formatter<Date> dateFormatter = FormatterFactory.getDateTimeFormatter(dc.getSessionControl());
			String name = dateFormatter.format(item.getLoginTime()) + 
				" (" + item.getRemoteId() + ")";
			if (link)
			{
				result = Base.getLink(ID, name, Item.SESSION, itemId, editLink);
			}
			else
			{
				result = HTML.encodeTags(name);
			}
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		if (result == null)
		{
			result = getLink(ID, null, itemId, denied, notFound, false, false);
		}
		return result;
	}
	
	/**
		Get the client application that was used to make the change in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getClient(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		Client item = null;
		Integer itemId = ch.getClientId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		try
		{
			item = ch.getClient(dc);
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		String result = getLink(ID, item, itemId, denied, notFound, link, editLink);
		return result;
	}

	
	/**
		Get the project that was active when the change was made in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getProject(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		Project item = null;
		Integer itemId = ch.getProjectId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		try
		{
			item = ch.getProject(dc);
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		String result = getLink(ID, item, itemId, denied, notFound, link, editLink);
		return result;
	}
	
	/**
		Get the plug-in that made the change in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getPlugin(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		PluginDefinition item = null;
		Integer itemId = ch.getPluginId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		try
		{
			item = ch.getPlugin(dc);
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		String result = getLink(ID, item, itemId, denied, notFound, link, editLink);
		return result;
	}
	
	/**
		Get the job that made the change in HTML format.
		@param dc A DbControl for database access
		@param ch The change history entry
		@param link TRUE to link the name to theview page
		@param editLink TRUE to enable CTRL/ALT/SHIFT-click for editing
		@return A HTML string
	*/
	public static String getJob(DbControl dc, ChangeHistory ch, boolean link, boolean editLink)
	{
		Job item = null;
		Integer itemId = ch.getJobId();
		boolean denied = false;
		boolean notFound = false;
		String ID = dc.getSessionControl().getId();
		try
		{
			item = ch.getJob(dc);
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		catch (ItemNotFoundException ex)
		{
			notFound = true;
		}
		String result = getLink(ID, item, itemId, denied, notFound, link, editLink);
		return result;
	}

	
	private static String getLink(String ID, Nameable item, Integer itemId, 
			boolean denied, boolean notFound, boolean link, boolean editLink)
	{
		String result = null;
		if (notFound)
		{
			result = "<i>- not found [id=" + itemId + "] -</i>";
		}
		else if (denied)
		{
			result = "<i>- denied -</i>";
		}
		else if (item == null)
		{
			result = "<i>- none -</i>";
		}
		else if (!link)
		{
			result = HTML.encodeTags(item.getName());
		}
		else
		{
			result = Base.getLinkedName(ID, item, denied, editLink);
		}
		return result;
	}
	
	
}
