/**
	$Id: EditUtil.java 5685 2011-08-04 10:58:23Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.edit;

import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.extensions.tabcontrol.TabAction;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.ExtensionPointBean;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.RethrowErrorHandlerFactory;

/**
	Utility class for handling edit-dialog related extensions. This class
	handler both the tab-control extension points in the edit-dialog and
	the on-save extension points when saving the changes.

	@author Nicklas
	@since 2.17
	@base.modified $Date: 2011-08-04 12:58:23 +0200 (Thu, 04 Aug 2011) $
*/
public class EditUtil
{

	
	/**
		An array containing all items that we are registering
		extension points for. The extension point id for an item
		is constructed as: EP_PREFIX_EDIT/EP_PREFIX_SAVE + Item.name().toLowerCase()
	*/
	public static final Item[] EDIT_ITEMS = new Item[] 
		{
			Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT, Item.TAG,
			Item.BIOPLATE, Item.BIOWELL, Item.BIOPLATETYPE, Item.BIOMATERIALLIST, Item.BIOMATERIALEVENT,
			Item.BIOPLATEEVENT, Item.BIOPLATEEVENTTYPE,
			Item.ARRAYDESIGN, Item.ARRAYBATCH, Item.ARRAYSLIDE,
			Item.PLATE, Item.PLATEEVENT, Item.WELL, Item.PLATETYPE, Item.PLATEEVENTTYPE,
			Item.PLATEGEOMETRY, Item.PLATEMAPPING,
			Item.DIRECTORY, Item.FILE, Item.FILESERVER, Item.PROJECT, Item.PERMISSIONTEMPLATE,
			Item.PHYSICALBIOASSAY, Item.DERIVEDBIOASSAY, Item.RAWBIOASSAY, 
			Item.FORMULA, Item.REPORTER, Item.REPORTERLIST,
			Item.EXPERIMENT, Item.TRANSFORMATION, Item.BIOASSAYSET, Item.BIOASSAY,
			Item.USER, Item.GROUP, Item.ROLE, Item.QUOTA, Item.SOFTWARE, Item.HARDWARE,
			Item.PROTOCOL, Item.CLIENT, Item.HELP, Item.NEWS,
			Item.PLUGINTYPE, Item.PLUGINDEFINITION, Item.PLUGINCONFIGURATION, Item.JOBAGENT,
			Item.PLATFORM, Item.PLATFORMVARIANT, Item.DATAFILETYPE,
			Item.QUOTATYPE, Item.MIMETYPE,
			Item.ANNOTATIONTYPECATEGORY, Item.ANNOTATIONTYPE,
			Item.QUANTITY, Item.UNIT, Item.EXTRAVALUETYPE, Item.REPORTERTYPE, Item.ITEMSUBTYPE
		};

	/**
		ID prefix for tabcontrol extension points on edit pages.
	*/
	public static String EP_PREFIX_EDIT = "net.sf.basedb.clients.web.tabcontrol.edit.";

	/**
		ID prefix for tabcontrol extension points on edit pages.
	*/
	public static String EP_PREFIX_SAVE = "net.sf.basedb.clients.web.onsave.";

	
	/**
		Use on-save extensions for a given gui context. This method
		will assemble all specific and generic extension points and
		then call {@link ExtensionsControl#useExtensions(JspContext, String...)}.
		
		@param jspContext The current jsp context
		@return An invoker instance
	*/
	public static ExtensionsInvoker useOnSaveExtensions(JspContext jspContext)
	{
		return useExtensions(EP_PREFIX_SAVE, jspContext);
	}

	/**
		Use edit-dialog extensions for a given gui context. This method
		will assemble all specific and generic extension points and
		then call {@link ExtensionsControl#useExtensions(JspContext, String...)}.
		
		@param jspContext The current jsp context
		@return An invoker instance
	*/
	public static ExtensionsInvoker useEditExtensions(JspContext jspContext)
	{
		return useExtensions(EP_PREFIX_EDIT, jspContext);
	}

	
	private static ExtensionsInvoker useExtensions(String prefix, JspContext jspContext)
	{
		String[] ep = new String[6]; // 6 is the maximum number of extension points
		int index = 0;
		GuiContext guiContext = jspContext.getGuiContext();
		Item itemType = guiContext.getItem();
		Class dataClass = itemType.getDataClass();
		ep[index++] = prefix + "all";
		if (dataClass != null)
		{
			if (OwnableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = prefix + "ownable";
			}
			if (ShareableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = prefix + "shareable";
			}
			if (RemovableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = prefix + "removable";
			}
			if (AnnotatableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = prefix + "annotatable";
			}
		}
		ep[index++] = prefix + itemType.name().toLowerCase();
		return ExtensionsControl.useExtensions(jspContext, ep);
	}

	/**
		Registers tabcontrol and on-save extension points with the given registry.
		Called at application startup.
	*/
	public static void registerExtensionPoints(Registry registry)
	{
		// Used for the on-save extensions
		RethrowErrorHandlerFactory ehf = new RethrowErrorHandlerFactory();
		
		registerEditTab(registry, EP_PREFIX_EDIT + "all", "Tab-control in all edit dialogs");
		registerEditTab(registry, EP_PREFIX_EDIT + "ownable", "Tab-control in edit dialogs for items that are owned by a user");
		registerEditTab(registry, EP_PREFIX_EDIT + "shareable", "Tab-control in edit dialogs for items that can be shared");
		registerEditTab(registry, EP_PREFIX_EDIT + "removable", "Tab-control in edit dialogs for items that can be put in the trashcan");
		registerEditTab(registry, EP_PREFIX_EDIT + "annotatable", "Tab-control in edit dialogs for annotatable items");

		registerOnSave(registry, EP_PREFIX_SAVE + "all", "On-save callback for all items", ehf);
		registerOnSave(registry, EP_PREFIX_SAVE + "ownable", "On-save callback for items that are owned by a user", ehf);
		registerOnSave(registry, EP_PREFIX_SAVE + "shareable", "On-save callback for items that can be shared", ehf);
		registerOnSave(registry, EP_PREFIX_SAVE + "removable", "On-save callback for items that can be put in the trashcan", ehf);
		registerOnSave(registry, EP_PREFIX_SAVE + "annotatable", "On-save callback for annotatable items", ehf);
	
		for (Item item : EDIT_ITEMS)
		{
			String suffix = item.name().toLowerCase();
			registerEditTab(registry, EP_PREFIX_EDIT + suffix, item + ": Tab-control in edit dialog");
			registerOnSave(registry, EP_PREFIX_SAVE + suffix, item + ": On-save callback", ehf);
		}
	}
	
	private static void registerEditTab(Registry registry, String id, String name)
	{
		ExtensionPointBean<TabAction> tabControl = new ExtensionPointBean<TabAction>();
		tabControl.setActionClass(TabAction.class);
		tabControl.setAllowRendererOverrider(false);
		tabControl.setId(id);
		tabControl.setName(name);
		tabControl.setDescription(
			"Extension point that adds a tab to the specified dialog. " +
			"Extensions should provide 'TabAction' instances with " +
			"information about the tab. Actions are always rendered by the " + 
			"TabControl taglib. This extension point supports " +
			"custom stylesheets and java-scripts."
			);
		registry.registerExtensionPoint(tabControl);
	}
	
	private static void registerOnSave(Registry registry, String id, String name, RethrowErrorHandlerFactory ehf)
	{
		ExtensionPointBean<OnSaveAction> onSave = new ExtensionPointBean<OnSaveAction>();
		onSave.setActionClass(OnSaveAction.class);
		onSave.setAllowRendererOverrider(false);
		onSave.setErrorHandlerFactory(ehf);
		onSave.setId(id);
		onSave.setName(name);
		onSave.setDescription(
			"Extension point that is invoked when the specified item is " +
			"saved after editing in the edit dialog. " +
			"Extensions should provide 'OnSaveAction' instances that " +
			"performs the desired action. The 'onSave' method is called after " +
			"all regular properties has been update. If the commit is successful " +
			"'onCommit' is called, otherwise 'onRollback'. The 'onSave' method " +
			"can throw an exception that will be forwarded to the gui. Do not throw " +
			"exception from the other 'onCommit/onRollback' since that will be " +
			"confusing."
			);
		registry.registerExtensionPoint(onSave);
	}

}
