/**
	$Id: MenuItemAction.java 4835 2009-03-24 08:22:28Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.menu;

import net.sf.basedb.util.extensions.Action;

/**
	An action for extensions to the menu system. The extension point
	will add menu items to the <code>Extensions</code> menu. 
	<p>
	There are three types of menu items:
	<ul>
	<li>A regular menu item, which is usually associated with some kind
		of action when the user clicks on it.
	<li>A separator item. It's just for visual appearance. It is 
		never associated with any action.
	<li>A submenu item, which opens up a submenu when the user
		selects it. A submenu must specify a {@link #getSubmenuId()}.
		Regular menu items can be put into the submenu item by specifying
		the same id for {@link #getMenuId()}.
	</ul>

	@author nicklas
	@version 2.7
	@base.modified $Date: 2009-03-24 09:22:28 +0100 (Tue, 24 Mar 2009) $
*/
public interface MenuItemAction
	extends Action
{

	/**
		The type of menu item.
	*/
	public MenuType getType();
	
	/**
		The ID of the menu where this menu item belongs, or null if
		it belongs to the top-level menu.
	*/
	public String getMenuId();
	
	/**
		The ID of a new submenu. This value is required when the menu item is
		a {@link MenuType#SUBMENU}, and is ignored otherwise.
	*/
	public String getSubmenuId();
	
	/**
		Get the title of the menu. 
	*/
	public String getTitle();
	
	/**
		A tooltip that is shown as a yellow popup when the 
		user points the mouse at the menu item.
		@return The tooltip or null to not display any tooltip
	*/
	public String getTooltip();
	
	/**
		Get an absolute reference to an image that will be displayed in the menu.
		It is recommended that the image is 9 pixels wide and 12 pixels high to
		line up with the icons used by the BASE core menus.	
		
		@return A reference to an image, or null if no image should
			be used
	*/
	public String getIcon();
	
	/**
		Extra CSS style attributes that will be added to the
		<code>style</code> attribute of the menu item div.
		@return Style attributes, or null
	*/
	public String getStyle();
	
	/**
		If the menu item should be enabled or disabled.
	*/
	public boolean isEnabled();
	
	/**
		If the menu item should be visible or not.
	*/
	public boolean isVisible();
	
	/**
		A javascript action that is executed when the user clicks the menu. 
		BASE has builtin scripts for opening a page in a popup window or opening
		a page in the main window:
		
		<ul>
		<li>To open a page in the main window use:
			<code>Menu.openUrl(url)</code>
			<p>
			The URL can be any valid URL. To access a JSP page on the BASE server
			you need to know the root path of the web application and the
			session ID for the logged in user. This information
			available in the JspContext passed to the factory class. If you,
			for example, want to open the reporters list page, you can
			create the URL like this:
			<pre class="code">
String url = jspContext.getRoot() + 
   "/views/reporters/index.jsp?ID=" + 
   jspContext.getSessionControl().getId();
</pre>
			
			It is also possible to use pure JavaScript to extract the root URL
			and session ID. Here is an example that opens the reporter list page:
			
			<pre class="code">
Menu.openUrl(getRoot()+'views/reporters/index.jsp?ID='+getSessionId());
</pre>

			Another possibility that works for the root path is to 
			use a factory that has enabled variable substitution. In this
			case you can use:

			<pre class="code">
Menu.openUrl('$ROOT$/views/reporters/index.jsp?ID='+getSessionId());
</pre>

		<li>To open a popup window use: 
			<code>Main.openPopup(url, windowName, width, height, options)</code>
			<p>
			The URL follows the same rules as the above. The <code>windowName</code>
			is a name to give to the popup window. Use only alphabetical characters (a-z).
			The size is controlled with the <code>width</code> and <code>height</code>
			options, given in pixels. BASE will automatically rescale the window
			if the user has choosen a bigger or smaller font size. Other 
			options can be specified in the <code>options</code> parameter. See
			documentation for the <code>window.open()</code> javascript function
			for valid syntax.
		</ul>
		
		@return The javascript action
	*/
	public String getOnClick();
	
	public static enum MenuType
	{
		/**
			This is a real menu item.
		*/
		MENUITEM,
		
		/**
			A menu item that opens a submenu. Menu items of this type must
			provide a unique value for {@link MenuItemAction#getSubmenuId()}.
			A regular menu item can be placed inside a submenu by specifying the
			same value for {@link MenuItemAction#getMenuId()}.
		*/
		SUBMENU,
		
		/**
			This is a separator line. All properties except 
			{@link MenuItemAction#isVisible()} and {@link MenuItemAction#getStyle()}
			are ignored.
		*/
		SEPARATOR;
	}
	
}
