/*
	$Id $

	Copyright (C) Nicklas Nordborg
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.servlet;

import java.io.IOException;
import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.News;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.util.MD5;

/**
	A servlet that generates a RSS feed from the public news on the BASE server.
	
	@author nicklas
	@since 2.17
	@base.modified $Date $
*/
public class RssNewsFeed 
	extends HttpServlet
{

	
	private static final long serialVersionUID = -4097480516166797186L;
	
	private static long lastUpdate;
	private static Date lastReset;
	private static boolean enabled;
	
	/**
		Reset the feed and force it to regenerate it's content.
		It is recommended that this method is called whenever
		a news items is added or modified. If not, the cache will
		expire sooner or later in any case.
	*/
	public static synchronized void reset()
	{
		lastUpdate = 0;
		lastReset = new Date();
		enabled = true;
	}

	/**
		Check if the rss servlet has been enabled or not.
	*/
	public static boolean isEnabled()
	{
		return enabled;
	}
	
	private String host;
	private String contextPath;
	private Charset utf8;
	private SessionControl sc;
	private byte[] cachedRssNews;
	
	public RssNewsFeed()
	{}	
	
	/*
		From the HttpServlet class
		--------------------------
	*/
	@Override
	public void init()
		throws ServletException
	{
		super.init();
		ServletContext context = getServletConfig().getServletContext();
		contextPath = context.getContextPath();
		host = Application.getHostName();
		utf8 = Charset.forName("UTF-8");
		reset();
	}

	@Override
	protected void doGet(HttpServletRequest request, HttpServletResponse response)
		throws ServletException, IOException
	{
		final String scheme = request.getScheme();
		final String serverName = request.getServerName();
		final int serverPort = request.getServerPort();
		final String rootUrl = scheme + "://" + serverName + ":" + serverPort + request.getContextPath();
		
		boolean debug = "debug".equals(request.getQueryString());
		
		// Refresh if debug or more than 10 minutes old
		byte[] feed = cachedRssNews;
		if (debug || lastUpdate + 600 * 1000 < System.currentTimeMillis())
		{
			String xml = getNewsFeed(rootUrl, debug);
			feed = xml.getBytes(utf8);
			// Only cache if not debugging
			if (!debug)
			{
				cachedRssNews = feed;
				lastUpdate = System.currentTimeMillis();
			}
		}
		
		response.setContentType("application/rss+xml");
		response.setContentLength(feed.length);
		response.setCharacterEncoding("UTF-8");
		response.getOutputStream().write(feed);
	}
	// -----------------------------

	/**
		Generate the XML string for the news feed.
	*/
	private String getNewsFeed(String rootUrl, boolean debug)
	{
		SessionControl sc = null;
		DbControl dc = null;
		List<News> news = null;
		String about = null;
		String adminName = null;
		String adminEmail = null;
		
		// Broadcast message from Adminstrate -> Broadcast message
		ServletContext ctx = getServletContext();
		String broadcastTitle = (String)ctx.getAttribute("broadcast.title");
		String broadcastMessage = (String)ctx.getAttribute("broadcast.message");

		// Load things from the database
		try
		{
			sc = Application.newSessionControl("net.sf.basedb.clients.web", null, null);
			dc = sc.newDbControl();
			
			about = sc.getClientDefaultSetting("server.about");
			adminName = sc.getClientDefaultSetting("server.admin.name");
			adminEmail = sc.getClientDefaultSetting("server.admin.email");

			ItemQuery<News> query = News.getQuery();
			query.order(Orders.desc(Hql.property("newsDate")));
			query.order(Orders.desc(Hql.property("id")));
			query.setCacheResult(true);
			news = new ArrayList<News>(query.list(dc));
		}
		finally
		{
			if (dc != null) dc.close();
			if (sc != null) sc.close();
		}
			
		// Generate the XML document
		String channelTitle = "BASE news from " + host + contextPath;
		SimpleDateFormat rfc822 = new SimpleDateFormat("EEE', 'dd' 'MMM' 'yyyy' 'HH:mm:ss' 'Z", Locale.US);

		// <rss version="2.0">
		Element rss = new Element("rss").setAttribute("version", "2.0");
		// <channel>
		Element channel = new Element("channel");
		// ..<title>.</title>
		channel.addContent(new Element("title").setText(channelTitle));
		// ..<link>.</link>
		channel.addContent(new Element("link").setText(rootUrl));
		// ..<description>.</description>
		channel.addContent(new Element("description").setText(about));
		// ..<webMaster>.</webMaster>
		if (adminEmail != null)
		{
			channel.addContent(new Element("webMaster").setText(adminEmail));
		}
		// ..<pubDate>.</pubDate>
		channel.addContent(new Element("pubDate").setText(rfc822.format(lastReset)));
		// ..<lastBuildDate>.</lastBuildDate>
		channel.addContent(new Element("lastBuildDate").setText(rfc822.format(new Date())));
		
		// ..<image>
		Element image = new Element("image");
		// ....<url>.</url>
		image.addContent(new Element("url").setText(rootUrl + "/images/baselogo.png"));
		// ....<title>.</title>
		image.addContent(new Element("title").setText(channelTitle));
		// ....<link>.</link>
		image.addContent(new Element("link").setText(rootUrl));
		// ..</image>
		channel.addContent(image);
		// </channel>
		rss.addContent(channel);
		
		// Create an <item> tag for the broadcast
		if (broadcastTitle != null)
		{
			String link = rootUrl + "/info/news.jsp#broadcast";
			// <item>
			Element item = new Element("item");
			// ..<link>.</link>
			item.addContent(new Element("link").setText(link));
			// ..<guid isPermaLink="false">.</guid>
			if (debug)
			{
				String guid = "broadcast-" + System.currentTimeMillis();
				item.addContent(new Element("guid").setAttribute("isPermaLink", "false").setText(guid));
			}
			else
			{
				String guid = MD5.getHashString(broadcastMessage);
				item.addContent(new Element("guid").setAttribute("isPermaLink", "false").setText(guid));
			}
			
			// ..<title>.</title>
			item.addContent(new Element("title").setText(broadcastTitle));
			// ..<description>.</description>
			item.addContent(new Element("description").setText(broadcastMessage));
			// ..<pubDate>.</pubDate>
			item.addContent(new Element("pubDate").setText(rfc822.format(new Date())));
			// </item>
			channel.addContent(item);
		}
		
		// Each News item goes into an <item> tag
		for (News n : news)
		{
			String link = rootUrl + "/info/news.jsp#" + n.getId() + "-" + n.getVersion();
			// <item>
			Element item = new Element("item");
			// ..<link>.</link>
			item.addContent(new Element("link").setText(link));
			// ..<guid isPermaLink="true">.</guid>
			if (debug)
			{
				String guid = n.getId() + "-" + n.getVersion() + "-" + System.currentTimeMillis();
				item.addContent(new Element("guid").setAttribute("isPermaLink", "false").setText(guid));
			}
			else
			{
				item.addContent(new Element("guid").setAttribute("isPermaLink", "true").setText(link));
			}
			// ..<title>.</title>
			item.addContent(new Element("title").setText(n.getName()));
			// ..<description>.</description>
			item.addContent(new Element("description").setText(n.getDescription()));
			// ..<pubDate>.</pubDate>
			item.addContent(new Element("pubDate").setText(rfc822.format(n.getStartDate())));
			// </item>
			channel.addContent(item);
		}
		
		Document dom = new Document(rss);
		String xml = new XMLOutputter(debug ? Format.getPrettyFormat() : Format.getCompactFormat()).outputString(dom);
		return xml;
	}
	

}
