/*
	$Id: TableColumn.java 5590 2011-03-16 11:48:47Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.web.taglib.table;

import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Type;
import net.sf.basedb.util.formatter.Formatter;

/**
   Column properties class used to define table columns.<br/>
   Example use in a jsp page:
   <pre>
&lt;%List&lt;TableColumn&gt; cols = new LinkedList&lt;TableColumn&gt;();
cols.add( new TableColumn("position", "POSITION", "int", "Position") );
cols.add( new TableColumn("ch1", "%1", "float", "Ch 1") );
cols.add( new TableColumn("ch2", "%2", "float", "Ch 2") );
cols.add( new TableColumn("externalId", "@externalId", "string", "External Id") );

for(TableColumn c : cols)
{%&gt;
	&lt;tbl:columndef 
		id="&lt;%=c.getId()%&gt;"
		property="&lt;%=c.getProperty()%&gt;"
		datatype="&lt;%=c.getDatatype()%&gt;"
		title="&lt;%=c.getTitle()%&gt;"
		sortable="&lt;%=c.getSortable()%&gt;" 
		filterable="&lt;%=c.getFilterable()%&gt;"
		exportable="&lt;%=c.getExportable()%&gt;"
		show="&lt;%=c.getShow()%&gt;" 
		formatter="&lt;%=c.getFormatter()%&gt;"
	/&gt;

&lt;%}%&gt;
   </pre>
   
  @author gregory, nicklas
	@version 2.0
	@base.modified $Date: 2011-03-16 12:48:47 +0100 (Wed, 16 Mar 2011) $
*/

public class TableColumn {
	private final String id;
	private final String property;
	private final String jepExpression;
	private final Type datatype; 
	private final String title;
	private final String description;
	private final boolean sortable;
	private final boolean filterable;
	private final boolean exportable;
	private final String show;
	private final Formula.AverageMethod averageMethod;
	private Formatter<?> formatter;
	
	public TableColumn(String id, String property, String jepExpression, Type datatype, String title, String description, String show,
		boolean sortable, boolean filterable, boolean exportable, Formula.AverageMethod averageMethod, Formatter<?> formatter)
	{		
		this.id = id;
		this.property = property;
		this.jepExpression = jepExpression;
		this.datatype = datatype; 
		this.title = title;
		this.description = description;
		this.show = show;
		this.sortable = sortable;
		this.filterable = filterable;
		this.exportable = exportable;
		this.averageMethod = averageMethod;
		this.formatter = formatter;
	}

	public Type getDatatype() 
	{
		return datatype;
	}

	public boolean getExportable() 
	{
		return exportable;
	}

	public boolean getFilterable() 
	{
		return filterable;
	}

	public String getId() 
	{
		return id;
	}

	public String getProperty() 
	{
		return property;
	}
	
	public String getJepExpression()
	{
		return jepExpression;
	}

	public String getShow() 
	{
		return show;
	}

	public boolean getSortable() 
	{
		return sortable;
	}

	public String getTitle() 
	{
		return title;
	}
	
	public String getDescription()
	{
		return description;
	}
	
	public boolean isAveragable()
	{
		return averageMethod != null && averageMethod != Formula.AverageMethod.NONE;
	}

	/**
		@return This table column's 
			{@link net.sf.basedb.core.Formula.AverageMethod} enum.
		@since 2.4
	*/
	public Formula.AverageMethod getAverageMethod()
	{
		return averageMethod;
	}
	
	/**
		@return This column's {@link net.sf.basedb.util.formatter.Formatter} object  
	 	@since 2.0.2
	*/
	public Formatter<?> getFormatter()
	{
		return formatter;
	}

	/**
		@param formatter Formatter to be set. 
	 	@since 2.0.2
	*/
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}
}
