/*
	$Id: DataFileType.java 5773 2011-09-29 13:14:56Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.info.DataFileTypeInfo;
import net.sf.basedb.info.ToTransferable;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.fileset.ValidationAction;

/**
	This class is used to represent types of files that are used
	to hold data instead of importing the data into into the database.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2011-09-29 15:14:56 +0200 (Thu, 29 Sep 2011) $
*/
public class DataFileType
	extends BasicItem<DataFileTypeData>
	implements Nameable, Removable, ToTransferable<DataFileTypeInfo>, Validatable
{
	
	/**
		The type of item represented by this class.
		@see Item#DATAFILETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.DATAFILETYPE;

	/**
		The external ID for the file type representing
		Affymetrix CDF files.
	*/
	public static final String AFFYMETRIX_CDF = "affymetrix.cdf";

	/**
		The external ID for the file type representing
		Affymetrix CEL files.
	*/
	public static final String AFFYMETRIX_CEL = "affymetrix.cel";

	/**
		The external ID for the file type representing
		a generic raw data file of any type.
	*/
	public static final String GENERIC_RAW_DATA = "generic.rawdata";

	/**
		The external ID for the file type representing
		a generic reporter map (array design layout) of any type.
	*/
	public static final String GENERIC_REPORTER_MAP = "generic.reportermap";

	/**
		The external ID for the file type representing
		a generic print map (array design layout) of any type.
	*/
	public static final String GENERIC_PRINT_MAP = "generic.printmap";

	/**
		The external ID for the file type representing
		generic analysed spot data.
		@since 2.8
	*/
	public static final String GENERIC_SPOT_DATA = "generic.spotdata";

	/**
		The external ID for the file type representing
		generic analysed extra value data.
		@since 2.8
	*/
	public static final String GENERIC_EXTRAVALUE_DATA = "generic.extravaluedata";

	/**
		The external ID for the file type representing
		a serial BASEFile.
		@since 2.11
	*/
	public static final String BASEFILE_SERIAL = "basefile.serial";
	
	/**
		The external ID for the file type representing
		a matrix BASEFile.
		@since 2.11
	*/
	public static final String BASEFILE_MATRIX = "basefile.matrix";
	
	/**
		The external ID for the file type representing a BFS
		metadata file. The metadata file may reference additional
		files.
		@since 2.15
	*/
	public static final String BFS_METADATA = "bfs.metadata";
	
	/**
		The external ID for the file type representing
		a MeV TDMS (Tab-Delimited Multiple Sample) file.
		@since 2.11
	*/
	public static final String MEV_TDMS = "mev.tdms";
	
	/**
		The external ID for the file type representing
		a SAM (Sequence Alignment/Map) file.
		@since 3.0
	*/
	public static final String SAM = "assembly.sam";

	/**
		The external ID for the file type representing
		a BAM (Binary Sequence Alignment/Map) file.
		@since 3.0
	*/
	public static final String BAM = "assembly.bam";

	/**
		The external ID for the file type representing
		scanned microarray image.
		@since 3.0
	*/
	public static final String MICROARRAY_IMAGE = "microarray.image";

	/**
		The external ID for the file type representing a GTF
		reference file that define the features on an array design
		that we are interested in.
		@since 3.0
	*/
	public static final String REF_SEQ_GTF = "refseq.gtf";
	
	/**
		The external ID for the file type representing a FPKM
		tracking file.
		http://cufflinks.cbcb.umd.edu/manual.html#tracking_format
		@since 3.0
	*/
	public static final String FPKM_TRACKING = "sequencing.fpkm_tracking";

	
	/**
		The maximum length of the file extension that can be stored in the database.
		@see #setExtension(String)
	*/
	public static final int MAX_EXTENSION_LENGTH = DataFileTypeData.MAX_EXTENSION_LENGTH;
	
	/**
		Create a new data file type.
		
		@param dc The DbControl to use for database access
		@param externalId The external ID of the new type, must be 
			unique for all file types
		@param itemType The type of item that files of this type can
			be attached to. The item must implement the {@link FileStoreEnabled}
			interface
		@return The new file set member type object
	*/
	public static DataFileType getNew(DbControl dc, String externalId, Item itemType)
	{
		DataFileType type = dc.newItem(DataFileType.class);
		type.setExternalId(externalId);
		type.setName("New data file type");
		type.setItemType(itemType);
		return type;
	}

	/**
		Get a <code>DataFileType</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>DataFileType</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static DataFileType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		DataFileType type = dc.loadItem(DataFileType.class, id);
		if (type == null) throw new ItemNotFoundException("DataFileType[id="+id+"]");
		return type;
	}
	
	/**
		Get a <code>DataFileType</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external ID of the item to load
		@return The <code>DataFileType</code> item.
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static DataFileType getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_DATAFILETYPE_FOR_EXTERNAL_ID");
		/*
			SELECT dft
			FROM DataFileTypeData dft
			WHERE dft.externalId = :externalId
		*/
		query.setString("externalId", externalId);
		DataFileType dft = dc.getItem(DataFileType.class, HibernateUtil.loadData(DataFileTypeData.class, query));
		if (dft == null) throw new ItemNotFoundException("DataFileType[externalId="+externalId+"]");
		return dft;
	}
	
	/**
		Get a query configured to retrieve <code>DataFileType</code> items.
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public static ItemQuery<DataFileType> getQuery()
		throws BaseException
	{
		return new ItemQuery<DataFileType>(DataFileType.class);
	}
	
	/**
		Get a query configured to retrieve <code>DataFileType</code>:s
		that can be used on a given item. This method will check the
		{@link FileStoreEnabled#getPlatform()} and {@link FileStoreEnabled#getVariant()}
		values and restrict the query to only return items that can be used for
		that platform. If the item isn't attached to any platform or variant
		all data file types for the {@link Identifiable#getType()} are returned.
		
		@param item The item to get data file types for, or null
			to return all data file types
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<DataFileType> getQuery(FileStoreEnabled item)
	{
		ItemQuery<DataFileType> query = null;
		if (item == null)
		{
			query = getQuery();
		}
		else
		{
			query = getQuery(item.getPlatform(), item.getVariant(), item.getType(), null);
		}
		return query;		
	}

	/**
		Get a query to retrieve <code>DataFileType</code>:s that can be used
		on specific platform/variant for a given item and/or generic file type.
		All parameters are optional. If all are null the result is the same
		as {@link #getQuery()}.
		
		@param platform The platform, or null to not restrict by platform (unless
			a variant is given)
		@param variant The platform variant, or null to not restrict by variant
		@param itemType The type of item it should be possible to
			attach file to, or null to not restrict by item
		@param genericType The generic file type, or null to not restrict
			by file type
		@return A query returning data file types matching the specified
			restrictions
		@since 2.9
	 */
	public static ItemQuery<DataFileType> getQuery(Platform platform, PlatformVariant variant,
		Item itemType, ItemSubtype genericType)
	{
		ItemQuery<DataFileType> query = getQuery();
		// Restrict on itemType
		if (itemType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemType"), 
					Expressions.integer(itemType.getValue())
				)
			);
		}
		
		// Restrict on generic file type
		if (genericType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("genericType"), 
					Hql.entity(genericType))
			);
		}
		
		// Restrict on platform/variant
		if (variant != null && platform == null) platform = variant.getPlatform();
		if (platform != null)
		{
			query.join(Hql.innerJoin("platforms", "pft"));
			query.restrict(Restrictions.eq(Hql.property("pft", "platform"), Hql.entity(platform)));
			if (variant != null)
			{
				query.restrict(Restrictions.or(
					Restrictions.eq(Hql.property("pft", "variant"), null),
					Restrictions.eq(Hql.property("pft", "variant"), Hql.entity(variant))
				));
			}
			else
			{
				query.restrict(Restrictions.eq(Hql.property("pft", "variant"), null));
			}
		}
		return query;
	}
	
	
	/**
		Creates a new filetype item from the given data.
	
		@param data the data.
	*/
	DataFileType(DataFileTypeData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
	From the BasicItem class
	-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A file has been tagged with this type in a file set
		<li>A platform is using this file type
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILESETMEMBERS_FOR_DATAFILETYPE", "count(*)");
		/*
			SELECT {1}
			FROM FileSetMemberData mbr
			WHERE mbr.dataFileType = :dataFileType
		*/
		query.setEntity("dataFileType", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATFORMFILETYPE_FOR_DATAFILETYPE", "count(*)");
			/*
				SELECT {1}
				FROM PlatformFileTypeData pft
				WHERE pft.dataFileType = :dataFileType
			*/
			query.setEntity("dataFileType", this.getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		return count > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link FileSetMember}:s using this file type
		<li>{@link PlatformFileType}:s using this file type
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILESETMEMBERS_FOR_DATAFILETYPE", "mbr.id");
			/*
				SELECT {1}
				FROM FileSetMemberData mbr
				WHERE mbr.dataFileType = :dataFileType
			*/
		query.setEntity("dataFileType", this.getData());
		addUsingItems(using, Item.FILESETMEMBER, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATFORMFILETYPE_FOR_DATAFILETYPE", "pft.id");
			/*
				SELECT {1}
				FROM PlatformFileTypeData pft
				WHERE pft.dataFileType = :dataFileType
			*/
		query.setEntity("dataFileType", this.getData());
		addUsingItems(using, Item.PLATFORMFILETYPE, query);
		return using;
	}

	/**
	 	@since 2.5
	 */
	public DataFileTypeInfo toTransferable(DataFileTypeInfo info)
	{
		super.toTransferable(info);
		info.setName(getName());
		info.setDescription(getDescription());
		info.setExtension(this.getExtension());
		info.setExternalId(this.getExternalId());
		return info;
	}
	// -------------------------------------------
	
	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = DataFileTypeData.MAX_EXTERNAL_ID_LENGTH;

	/**
		Get the external id for the data file type.
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	private void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		getData().setExternalId(StringUtil.setNullableString(
			externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
		Get the generic file type for this data file type.
		@return A file subtype item
		@since 3.0
	*/
	public ItemSubtype getGenericType()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getGenericType());
	}
	
	/**
		Set the generic file type of this item. Null is not allowed.
		@param genericType The generic file type
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@since 3.0
	*/
	public void setGenericType(ItemSubtype genericType)
	{
		checkPermission(Permission.WRITE);
		if (genericType != null) genericType.checkItemType(Item.FILE);
		getData().setGenericType(genericType == null ? null : genericType.getData());
	}
	
	/**
		Get the common file extension for data files of this type.
		@return A string or null if there is no common file extension
	*/
	public String getExtension()
	{
		return getData().getExtension();
	}

	/**
		Set the most common gile extension for data file of this
		type.
		@param extension The extension or null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_EXTENSION_LENGTH}.
	*/
	public void setExtension(String extension)
	{
		checkPermission(Permission.WRITE);
		getData().setExtension(StringUtil.setNullableString(extension, "extension", MAX_EXTENSION_LENGTH));
	}

	private void setItemType(Item itemType)
	{
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		if (!FileStoreEnabled.class.isAssignableFrom(itemType.getItemClass()))
		{
			throw new InvalidDataException(itemType + " isn't a subclass of FileStoreEnabled");
		}
		getData().setItemType(itemType.getValue());
	}
	
	/**
		Get the type of item files of this type can be used on. The
		item should implement the {@link FileStoreEnabled} interface.
		@return An <code>Item</code> object
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}

	/**
		Get a query returning the item subtypes that files of this type
		can be used on. The query should only be used if the main item
		type {@link #getItemType()} is a {@link Subtypable} item.
		@return A query
		@since 3.0
	*/
	public ItemQuery<ItemSubtype> getItemSubtypes()
	{
		ItemQuery<ItemSubtype> query = ItemSubtype.getQuery(null);
		query.joinPermanent(Hql.innerJoin("dataFileTypes", Item.DATAFILETYPE.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.DATAFILETYPE.getAlias()), 
				Hql.entity(this)
		));
		return query;
	}
	
	/**
		Check if there is at least one extension that claims to support 
		validation of files of this type. The extensions are not actually
		invoked, only prepared.
		@since 3.0
	*/
	@SuppressWarnings("unchecked")
	public boolean hasActiveValidator(DbControl dc)
	{
		ExtensionsManager manager = Application.getExtensionsManager();
		Registry registry = manager.getRegistry();
		ClientContext context = new ClientContext(dc, this);
		ExtensionsInvoker<ValidationAction> invoker = 
					(ExtensionsInvoker<ValidationAction>)registry.useExtensions(context, manager.getSettings(), "net.sf.basedb.core.filehandler.validator");
		return invoker.getNumExtensions() > 0;
	}
	
	/**
		Get a query that returns all platforms/variants that are using this
		file type.
		@return A query object
	*/
	public ItemQuery<PlatformFileType> getPlatforms()
	{
		ItemQuery<PlatformFileType> query = PlatformFileType.getQuery(null, null, false);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("dataFileType"),
				Hql.entity(this)
			)
		);
		return query;
	}
	
}
