/*
	$Id: Aggregations.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;

/**
	A factory class to create aggregate expressions.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/
public class Aggregations
{

	/**
		Calculates the number of values for an expression: new expression = COUNT(e)
		@param e The expression to count, or null to count all rows (ie. COUNT(*))
		@param distinct If only distinct values should be counted
		@return The new Expression
	*/
	public static Expression count(Expression e, boolean distinct)
	{
		return new CountExpression(e, distinct);
	}

	/**
		Calculates the geometric mean of all values for an expression: new expression = 
		EXP(AVG(LN(e))). Note that this method can handle values less then or equal with
		0 although the mathematical definition of geometric mean doesn't allow it. 
		This is caused by the way that sql handle null values in aggregated operations. 
		The geometric mean of the numbers [2, -5, 8] is therefor 4 (because ln(-5) is 
		undefined, i.e. null, it is ignored in the avg operation). If you ask a 
		mathematician he would say that the geometic mean is undefined for those numbers.
		
		@param e The expression to use in the calculation
		@return The new Expression
		@throws InvalidDataException If the expression is null
		@since 2.4
	*/
	public static Expression geometricMean(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new GeometricMeanExpression(e);
	}

	/**
		Calculates the maximum value of an expression: new expression = MAX(e)
		@param e The expression to use in the calculation
		@return The new Expression
		@throws InvalidDataException If the expression is null
	*/
	public static Expression max(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new MaxExpression(e);
	}

	/**
		Calculates the minimum value of an expression: new expression = MIN(e)
		@param e The expression to use in the calculation
		@return The new Expression
		@throws InvalidDataException If the expression is null
	*/
	public static Expression min(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new MinExpression(e);
	}

	/**
		Calculates the arithmetic mean of all values for an expression: new expression = AVG(e)
		@param e The expression to use in the calculation
		@return The new Expression
		@throws InvalidDataException If the expression is null
	*/
	public static Expression mean(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new MeanExpression(e);
	}

	/**
		Calculates the sum of all values of an expression: new expression = SUM(e)
		@param e The expression to use in the calculation
		@return The new Expression
		@throws InvalidDataException If the expression is null
	*/
	public static Expression sum(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new SumExpression(e);
	}
	
	/**
	 	Calculates the quadratic mean, also called Root Mean Square, for an expression: 
	 	new expression = sqrt(AVG(e*e))
	 	@param e The expression to use in the calculation
	 	@return The new Expression
	 	@throws InvalidUseOfNullException If the expression is null
	 	@since 2.8
	 */
	public static Expression quadraticMean(Expression e)
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new QuadraticMeanExpression(e);
	}
}
