/*
	$Id: IntensityCalculatorUtil.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.DataQuery;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.PositionBatcher;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.MappingBatcher;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.DataResultIterator;
import net.sf.basedb.core.DynamicRawDataQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.jep.Jep;
import net.sf.basedb.util.jep.RawFunction;
import net.sf.basedb.util.jep.MeanFunction;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.sql.SQLException;

import org.nfunk.jep.JEP;

/**
	This is a utility class for plugins that calculate intensities
	from raw spot data.
	
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
	@author Nicklas
	@version 2.0
*/
public class IntensityCalculatorUtil
{

	public static boolean checkSameArrayDesign(Collection<RawBioAssay> rawBioAssays, boolean throwException)
	{
		ArrayDesign ad = null;
		for (RawBioAssay rba : rawBioAssays)
		{
			if (ad == null)
			{
				ad = rba.getArrayDesign();
				if (ad == null)
				{
					if (throwException)
					{
						throw new InvalidDataException("Raw bioassay doesn't have an array design: "+rba);
					}
					return false;
				}
				else if (!ad.hasFeatures())
				{
					if (throwException)
					{
						throw new InvalidDataException("Array design doesn't have any features: "+ad);
					}
					return false;
				}
			}
			else
			{
				if (!ad.equals(rba.getArrayDesign()))
				{
					if (throwException)
					{
						throw new InvalidDataException("Raw bioassay uses a different array design: " +rba);
					}
					return false;
				}
			}
		}	
		return true;
	}
	
	/**
		Create a root bioassayset for a given experiment and set of raw bioassays.
		The raw bioassays must be part of the experiment. The intensities for
		the bioassays in the new bioassayset are calculated by adding Query
		API {@link Expression} objects to a {@link DynamicRawDataQuery}.
		This method is very fast since most of the work is done on the
		database only.
		<p>
		This method can only be used if all raw bioassays are connected
		to the same {@link ArrayDesign} (which must have features), since it
		doesn't supports remapping of the position numbers.
		
		@param dc The <code>DbControl</code> object to use for database acecss
		@param experiment The experiment where the new bioassayset should be placed
		@param rawBioAssays The raw bioassays to create bioassays for, or null
			to use all raw bioassays
		@param job The job that the transformation should be linked to, or
			null if this method isn't called by a job
		@param intensities An array containing the expressions for 
			calculating the intensities, one for each channel
	 	@param progress Optional progress reporter if the caller is interested in the progress 
	 	@return A {@link net.sf.basedb.core.BioAssaySet} object. 
		@throws InvalidDataException If any of the required parameters is null
			or if the raw bioassays arn't using the same array design or
			if the number of intensity expressions doesn't match the number of channels
	 	@throws BaseException If anything else goes wrong in the process. 
	*/
	public static BioAssaySet createRootBioAssaySet(DbControl dc, Experiment experiment, 
		Collection<RawBioAssay> rawBioAssays, Job job, Expression[] intensities, ProgressReporter progress)
		throws InvalidDataException, BaseException
	{

		if (experiment == null) throw new InvalidUseOfNullException("experiment");
		if (intensities == null) throw new InvalidUseOfNullException("intensities");
	
		if (intensities.length != experiment.getRawDataType().getChannels())
		{
			throw new InvalidDataException("Wrong number of intensity expressions: "+intensities.length+
				"; expected "+experiment.getRawDataType().getChannels());
		}
		
		checkSameArrayDesign(rawBioAssays, true);
		ThreadSignalHandler.checkInterrupted();
		
		// Load raw bioassays
		if (rawBioAssays == null || rawBioAssays.size() == 0)
		{
			rawBioAssays = experiment.getRawBioAssays().list(dc);
		}

		// Create Transformation
		Transformation t = experiment.newTransformation(job, rawBioAssays);
		dc.saveItem(t);

		// Create the root BioAssaySet in a new data cube and layer
		BioAssaySet root = t.newProduct("new", "new", false);
		root.setName("New root bioassayset");
		dc.saveItem(root);

		// Create the batchers we need
		PositionBatcher posBatcher = root.getPositionBatcher();
		SpotBatcher spotBatcher = root.getSpotBatcher();
		MappingBatcher rawBatcher = root.getMappingBatcher();

		int totalRawBioAssays = rawBioAssays.size();
		int rawBioAssaysDone = 0;
		// Spot + raw mapping for each bioassay + reporter mapping
		float stepSize = 90f / (float)(totalRawBioAssays * 2 + 1);
		float percent = 5;
		
		boolean first = true;
		for (RawBioAssay rba : rawBioAssays)
		{
			// Create a new BioAssay for each raw bioassay
			BioAssay ba = root.newRootBioAssay(Collections.singleton(rba));
			ba.setName(rba.getName());
			dc.saveItem(ba);

			// The query for this raw bioassay
			DynamicRawDataQuery query = rba.getDynamicRawData();
			Select column = Selects.expression(Expressions.integer(ba.getDataCubeColumnNo()), VirtualColumn.COLUMN.getName());
			Select position = Dynamic.selectRawData("position");

			// Map reporters to positions
			if (first)
			{
				if (progress != null)
				{
					progress.display((int)percent, "Mapping reporters for " + rba.getNumDbSpots() + " spots");
					percent += stepSize;
				}
				query.select(position);
				query.select(Dynamic.selectRawData("reporter"));
				posBatcher.insert(query);
				query.reset();
				first = false;
				ThreadSignalHandler.checkInterrupted();
			}

			// Calculate spot intensities
			if (progress != null)
			{
				progress.display((int)percent, "Calculating spot intensities; "+rawBioAssaysDone+
					" of " + totalRawBioAssays + " raw bioassay(s) done...");
				percent += stepSize;
			}
			query.select(column);
			query.select(position);
			for (int i = 0; i < intensities.length; ++i)
			{
				query.select(Selects.expression(intensities[i], "ch" + i));
			}
			spotBatcher.insert(query);
			query.reset();
			ThreadSignalHandler.checkInterrupted();
			
			// Map spot data to raw data
			if (progress != null)
			{
				progress.display((int)percent, "Mapping raw data; "+rawBioAssaysDone+" of " + 
					totalRawBioAssays + " raw bioassay(s) done...");
				percent += stepSize;
			}
			query.select(column);
			query.select(position);
			query.select(Dynamic.selectRawData("id"));
			rawBatcher.insert(query);
			query.reset();
			rawBioAssaysDone++;
			ThreadSignalHandler.checkInterrupted();
		}
		
		// Close batchers and clean up
		posBatcher.close();
		spotBatcher.close();
		rawBatcher.close();

		// Return  
		return root;
	}

	/**
		Create a root bioassayset for a given experiment and set of raw bioassays.
		The raw bioassays must be part of the experiment. The intensities for the
		bioassays in the new bioassayset are calculated by a {@link IntensityCalculator}
		object. This method is not as fast as the other <code>createRootBioAssay</code>
		method since all raw data must be loaded into memory and the inserted into
		the database one row at a time.
		<p>
		On the other hand, this method supports remapping of the position numbers
		so it is not necessary for all raw bioassays to use the same array design.

		@param dc The <code>DbControl</code> object to use for database acecss
		@param experiment The experiment where the new bioassayset should be placed
		@param rawBioAssays The raw bioassays to create bioassays for, or null
			to use all raw bioassays
		@param job The job that the transformation should be linked to, or
			null if this method isn't called by a job
		@param iCalc An object implementing the <code>IntensityCalculator</code>
			interface
		@param progress A progress reporter object, or null if the caller isn't
			interested in the progress
	 	@return A {@link net.sf.basedb.core.BioAssaySet} object 
	 	@throws InvalidDataException If any of the required parameters is null. 
	 	@throws BaseException If there is another error.
	*/
	public static BioAssaySet createRootBioAssaySet(DbControl dc, Experiment experiment, 
		Collection<RawBioAssay> rawBioAssays, Job job, IntensityCalculator iCalc, ProgressReporter progress)
		throws InvalidDataException, BaseException
	{

		if (experiment == null) throw new InvalidUseOfNullException("experiment");
		if (iCalc == null) throw new InvalidUseOfNullException("iCalc");
	
		// Create Transformation
		Transformation t = experiment.newTransformation(job, rawBioAssays);
		dc.saveItem(t);
		
		if (progress != null) progress.display(5, "Initialising...");

		// Load raw bioassays
		if (rawBioAssays == null || rawBioAssays.size() == 0)
		{
			rawBioAssays = experiment.getRawBioAssays().list(dc);
		}

		// Create the root BioAssaySet in a new data cube and layer
		BioAssaySet root = t.newProduct("new", "new", false);
		root.setName("New root bioassayset");
		dc.saveItem(root);

		// Create the batchers we need
		PositionBatcher posBatcher = root.getPositionBatcher();
		SpotBatcher spotBatcher = root.getSpotBatcher();
		MappingBatcher rawBatcher = root.getMappingBatcher();

		// A cache holding already seen position --> reporter mappings
		Map<Integer, Integer> positionCache = new HashMap<Integer, Integer>();
		// A cache holding already seen feature->position mappings
		Map<Integer, Integer> featureCache = new HashMap<Integer, Integer>();
		
		// The max position number in the raw bioassays -- only calculated if needed (see below)
		int maxPosition = 0;
		
		long spotsDone = 0; // Number of calculated spots
		int totalSpots = 0; // Total number of spots
		int interval = 0; // Interval between reporter to the progress reporter
		if (progress != null) 
		{
			for (RawBioAssay rba : rawBioAssays)
			{
				totalSpots += rba.getNumDbSpots();
			}
			interval = totalSpots / 50;
			if (interval < 10) interval = 10;
			progress.display(10, "Calculating spot intensities (0 done)...");
		}
		
		for (RawBioAssay rba : rawBioAssays)
		{
			// Create a new BioAssay for each raw bioassay
			BioAssay ba = root.newRootBioAssay(Collections.singleton(rba));
			ba.setName(rba.getName());
			dc.saveItem(ba);
		
			// This is the column coordinate in the datacube
			short column = ba.getDataCubeColumnNo();
		
			// Load raw data for current raw bioassay
			DataQuery<RawData> rawQuery = rba.getRawData();
			DataResultIterator<RawData> rawData = rawQuery.iterate(dc);
			ThreadSignalHandler.checkInterrupted();
			while (rawData.hasNext())
			{
				ThreadSignalHandler.checkInterrupted();
				RawData rd = rawData.next();
				int position = rd.getPosition();
				ReporterData reporter = rd.getReporter();
				FeatureData feature = rd.getFeature();
				
				// Calculate intensities
				float[] intensities = iCalc.calculateIntensities(rd);
				
				if (intensities != null)
				{
					// If we have a feature check the feature cache if a position has
					// already been assigned to it
					Integer cachedFeaturePosition = feature == null ?
						null : featureCache.get(feature.getId());
					if (cachedFeaturePosition != null)
					{
						position = cachedFeaturePosition;
					}
					else
					{
						// We have no feature or the feature has no position. 
						// Check if the position has same reporter
						if (positionCache.containsKey(position))
						{
							Integer cachedReporterId = positionCache.get(position);
							Integer reporterId = reporter == null ? null : reporter.getId();
							if (reporterId == null && cachedReporterId != null || 
								reporterId != null && !reporterId.equals(cachedReporterId))
							{
								// Reporters don't match, we need a new position
								if (maxPosition == 0)
								{
									// We must find the maximum position among all raw bioassays unless we havn't done that already
									try
									{
										for (RawBioAssay rbaMax : rawBioAssays)
										{
											DynamicRawDataQuery query = rbaMax.getDynamicRawData();
											query.select(Selects.expression(Aggregations.max(Dynamic.rawData("position")), "maxPosition"));
											DynamicResultIterator result = query.iterate(dc);
											int thisMaxPosition = result.next().getInt(1);
											if (thisMaxPosition > maxPosition) maxPosition = thisMaxPosition;
											result.close();
										}
									}
									catch (SQLException ex)
									{
										throw new BaseException(ex);
									}
								}
								maxPosition++;
								position = maxPosition;
								posBatcher.insert(position, reporter);
								positionCache.put(position, reporterId);
								if (feature != null) featureCache.put(feature.getId(), position);
							}
						}
						else
						{
							// This position is not used, map it to the current reporter
							posBatcher.insert(position, reporter);
							positionCache.put(position, reporter == null ? null : reporter.getId());
							if (feature != null) featureCache.put(feature.getId(), position);
						}
					}
					
					// Insert mapping to raw data spot
					rawBatcher.insert(column, position, rd);
					// Insert intensities
					spotBatcher.insert(column, position, intensities);
				}
				if (progress != null)
				{
					spotsDone++;
					if (spotsDone % interval == 0)
					{
						int percent = 10 + (int)((90L * spotsDone) / totalSpots);
						progress.display(percent, "Calculating spot intensities ("+spotsDone+" done)...");
					}
				}
			}
		}
		if (progress != null) progress.display(100, "Calculating spot intensities ("+totalSpots+" done)...\n");
	
		// Close batchers and clean up
		positionCache.clear();
		posBatcher.close();
		spotBatcher.close();
		rawBatcher.close();
		
		// Return  
		return root;
	}
	
	/**
		Create an intensity calculator from mathematical expressions in strings. The expressions
		may contain most regular mathematical operators, including <code>+, -, / , *, ln</code>.
		Use the name of a raw() function to include the value of that property, or the
		mean() function to include the mean of a property (over the whole data set). Examples
		of valid expressions are (assuming that the properties has been defined for the
		raw data type in question; see {@link RawDataType#getProperties()}):
		<pre class="code">
raw('ch1FgMean') - raw('ch1BgMean')
raw('ch1FgMedian') - mean('ch1BgMedian')
</pre>
		
		@param dc The DbControl used for database access
		@param rawDataType The raw data type to create the intensity calculator for. Null is not allowed.
		@param jepFormulas The mathematical expressions in a form supported by JEP,
			one expression is required for each channel. Null is not allowed
		@return An intensity calculator object
	 	@throws InvalidDataException If any of the required arguments is null or
	 		if the formulas length does not match the number of channels. 
	 	@throws BaseException If there is another error.
		@see RawDataType#getIntensityFormulas()
		@see net.sf.basedb.core.IntensityFormula
		@see net.sf.basedb.util.jep.Jep
	*/
	public static IntensityCalculator createJepIntensityCalculator(DbControl dc, RawDataType rawDataType, String... jepFormulas)
		throws InvalidDataException, BaseException
	{
		if (rawDataType == null) throw new InvalidUseOfNullException("rawDataType");
		if (jepFormulas == null || jepFormulas.length == 0) throw new InvalidUseOfNullException("jepFormulas");
	
		int channels = rawDataType.getChannels();
		if (channels != jepFormulas.length) 
		{
			throw new InvalidDataException("Invalid number of formulas, expected " + channels + "; found "+jepFormulas.length);
		}
		JEP[] jeps = new JEP[channels];
		RawFunction raw = new RawFunction(dc, rawDataType, true);
		MeanFunction mean = new MeanFunction(dc);
		for (int i = 0; i < channels; ++i)
		{
			if (jepFormulas[i] == null)
			{
				throw new InvalidUseOfNullException("jepFormulas["+i+"]");
			}
			jeps[i] = Jep.newJep(jepFormulas[i], raw, mean);
		}
		return new JepIntensityCalculatorImpl(jeps, raw, mean);
	}

	public static Expression[] createJepExpressions(DbControl dc, RawDataType rawDataType, String[] formulas)
	{
		Expression[] expressions = new Expression[formulas.length];
		RawFunction raw = new RawFunction(dc, rawDataType, true);
		MeanFunction mean = new MeanFunction(dc);
		for (int i = 0; i < formulas.length; ++i)
		{
			expressions[i] = Jep.formulaToExpression(formulas[i], raw, mean);
		}
		return expressions;
	}
	
	/**
		An implementation of the {@link IntensityCalculator} interface using
		JEP expressions.
		@see net.sf.basedb.util.jep.Jep
	*/
	private static class JepIntensityCalculatorImpl
		implements IntensityCalculator
	{
		private final JEP[] jeps;
		private final RawFunction raw;
		private final MeanFunction mean;
		
		private JepIntensityCalculatorImpl(JEP[] jeps, RawFunction raw, MeanFunction mean)
		{
			this.jeps = jeps;
			this.raw = raw;
			this.mean = mean;
		}
	
		/*
			From the IntensityCalculator interface
			-------------------------------------------
		*/
		public float[] calculateIntensities(RawData rawData)
		{
			float[] intensities = new float[jeps.length];
			for (int i = 0; i < jeps.length; ++i)
			{
				JEP jep = jeps[i];
				raw.setRawData(rawData);
				mean.setRawData(rawData);
				intensities[i] = (float)jep.getValue();
			}
			return intensities;
		}
		// -------------------------------------------
	}
}
