/**
	$Id: ChildrenTransformer.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.biomaterial;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;
import net.sf.basedb.util.collections.CollectionTransformer;

/**
	Collection transformer that given a collection of biomaterials
	finds all children of the same type.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
*/
public class ChildrenTransformer<I extends MeasuredBioMaterial>
	implements CollectionTransformer<I, I>
{

	private final DbControl dc;
	private final ItemQuery<I> query;
	private final boolean copySources;
	
	public ChildrenTransformer(DbControl dc, boolean copySources, ItemQuery<I> query)
	{
		this.dc = dc;
		this.copySources = copySources;
		this.query = query;
		query.include(Include.ALL);
		query.join(Hql.innerJoin("creationEvent", "ce"));
		query.join(Hql.innerJoin("ce", "sources", "sbm"));
		query.join(Hql.innerJoin("sbm", "bioMaterial", "bm"));
		query.restrict(
			Restrictions.in(
				Hql.alias("bm"), 
				Expressions.parameter("sources")
			)
		);
	}
	
	/*
		From the CollectionTransformer interface
		----------------------------------------
	*/
	/**
		
	*/
	@Override
	public int transform(Collection<? extends I> source,
			Collection<? super I> destination)
	{
		// Copy source ID:s so we can use them as a parameter in the query
		Set<Integer> sourceIds = new HashSet<Integer>();
		BasicItemToIdTransformer idTransformer = new BasicItemToIdTransformer();
		idTransformer.transform(source, sourceIds);
		int originalSize = destination.size();
		
		// Copy source biomaterials
		if (copySources) destination.addAll(source);
		
		// We need to loop so we can move down as many levels as we need
		while (sourceIds.size() > 0)
		{
			query.setParameter("sources", sourceIds, Type.INT);
			List<I> children = query.list(dc);
			// Add the children to the destination. If they all already exists we break to 
			// loop (to avoid problematic circular references)
			if (!destination.addAll(children)) break;
			sourceIds.clear();
			// Next iteration will find the children of the children, and so on...
			idTransformer.transform(children, sourceIds);
		}
		return destination.size() - originalSize;
	}
	// --------------------------------------------------

}
