/**
	$Id: ExtensionPointContext.java 5488 2010-11-15 12:32:35Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import net.sf.basedb.util.extensions.Registry.RegisteredExtensionPoint;

/**
	Invokation context for an extension point.

	@author nicklas
	@version 2.7
	@base.modified $Date: 2010-11-15 13:32:35 +0100 (Mon, 15 Nov 2010) $
*/
class ExtensionPointContext<A extends Action>
	extends InvokationContext<A>
{
	
	private final ClientContext clientContext;
	private final Registry registry;
	private final RegisteredExtensionPoint<A> extensionPoint;
	private final ErrorHandlerFactory<? super A> errorHandlerFactory;
	private Renderer<? super A> renderer;
	private boolean hasCreatedRenderer;
	private ErrorHandler<? super A> errorHandler;
	
	ExtensionPointContext(Registry registry, ClientContext clientContext, 
			RegisteredExtensionPoint<A> extensionPoint,
			ErrorHandlerFactory<? super A> errorHandlerFactory)
	{
		this.registry = registry;
		this.clientContext = clientContext;
		this.extensionPoint = extensionPoint;
		this.errorHandlerFactory = errorHandlerFactory;
	}
	
	/*
		From the InvokationContext class
		--------------------------------
	*/
	@Override
	public ClientContext getClientContext()
	{
		return clientContext;
	}
	
	@Override
	public ExtensionPoint<? super A> getExtensionPoint()
	{
		return extensionPoint;
	}
	
	/**
		Always null, since this is an extension point context.
	*/
	@Override
	public Extension<A> getExtension()
	{
		return null;
	}
	
	/**
		Always null, since this is an extension point context.
	*/
	@Override
	public Object getAttribute(String name)
	{
		return null;
	}

	@Override
	protected Registry getRegistry()
	{
		return registry;
	}
	
	@Override
	protected void prepareRendererFactory()
	{
		RendererFactory<? super A> factory = extensionPoint.getRendererFactory();
		if (factory != null) 
		{
			try
			{
				factory.prepareContext(this);
			}
			catch (Throwable t)
			{
				handleError("Error preparing renderer factory for extension point '" +
						extensionPoint.getId() + "': " + factory, t);
			}
		}
	}
	
	/**
		If the extension point defines a renderer factory, creates
		a new renderer unless one already exists.
		@return A renderer, or null if no factory has been defined
	*/
	@Override
	protected Renderer<? super A> getRenderer()
	{
		if (!hasCreatedRenderer)
		{
			RendererFactory<? super A> factory = extensionPoint.getRendererFactory();
			if (factory != null)
			{
				try
				{
					renderer = factory.getRenderer(this);
				}
				catch (Throwable t)
				{
					handleError("Error creating default renderer for for extension point '" +
							extensionPoint.getId() + "': " + factory, t);
				}
			}
			hasCreatedRenderer = true;
		}
		return renderer;
	}

	/**
		Always null, since this is an extension point context.
	*/
	@Override
	protected A[] getActions()
	{
		return null;
	}
	
	@Override
	protected void setError(A action, Throwable t)
	{
		extensionPoint.setError(action, t);
	}
	
	@Override
	protected void clearError()
	{
		extensionPoint.clearError();
	}
	
	@Override
	protected ErrorHandler<? super A> getErrorHandler()
	{
		if (errorHandler == null)
		{
			errorHandler = errorHandlerFactory.getErrorHandler(this);
		}
		return errorHandler;
	}
	// ----------------------------------------

	private void handleError(String message, Throwable t)
	{
		getErrorHandler().handleError(null, null, message, t);
	}

	
}