/**
	$Id: BeanActionFactory.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.debug;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	Generic action factory class that can generate instances of
	actions if there exists a bean-like implementation of the
	action class.
	<p>
	
	Parameters in the XML file will be set on the bean if there is 
	a setter method with the same name as the parameter tag, 
	following the usual syntax rules. Eg. the parameter value 
	<code>&lt;icon&gt;/images/icon.gif&lt;/icon&gt; will
	be passed to the <code>setIcon(String)</code> method on the
	bean. If there is no such method the factory will check if 
	there is a <code>getIcon()</code> or <code>isIcon()</code> 
	method with a return type other than <code>String</code>. If so,
	the factory will try a second time to find a <code>setIcon()</code>
	method with the same parameter type. Methods corresponding to 
	parameters that have not been set in the XML file are never
	called.
	<p>
	
	The factory can convert the string values to other data types for the 
	following classes: int, long, float, double and boolean. It can handle 
	both primitive and wrapper classes. Example:
	<code>&lt;enabled&gt;1&lt;/enabled&gt; will return <code>true</code> if
	<code>isEnabled()</code> is called on the bean.
	<p>
	
	<b>Parameters supported by this factory</b><br>
	This factory supports all parameters. Some parameters may trigger special
	actions:
	
	<ul>
	<li>beanClass: The class name of a class that implements the
		extension points action class.
	<li>disabled: If <code>true</code> the {@link #prepareContext(InvokationContext)} 
		method returns false
	</ul>

	@author nicklas
	@version 2.7
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
*/
public class BeanActionFactory
	implements ActionFactory<Action>
{

	private String beanClassName;
	private boolean disabled;
	private Action bean;
	private Action[] actions;
	private Map<String, String> parameters;
	private volatile boolean initialised;
	
	/**
		Creates a new proxy factory.
	*/
	public BeanActionFactory()
	{}

	/*
		From the ActionFactory interface
		--------------------------------
	*/
	@Override
	public boolean prepareContext(InvokationContext<? super Action> context)
	{
		if (disabled) return false;
		if (!initialised)
		{
			initBean(context.getExtensionPoint().getActionClass());
		}
		return bean != null;
	}

	@Override
	public Action[] getActions(InvokationContext context)
	{
		return disabled ? null : actions;
	}
	// -----------------------------------
	
	/**
		Set generic parameters.
		@param name The name of the parameter
		@param value The value of the parameter
	*/
	@PathSetter
	@VariableSetter
	public void setParameter(String name, String value)
	{
		if (parameters == null) 
		{
			parameters = new HashMap<String, String>();
		}
		parameters.put(name, value);
	}
	
	/**
		Set the name of the class to use as a bean. The bean
		must implement the action class of the extension point.
		@param beanClass
	*/
	public void setBeanClass(String beanClass)
	{
		this.beanClassName = beanClass;
	}
	
	/**
		Sets the disabled/enabled status of this factory.
		@param disabled A string that is parsed to a boolean
			by {@link Values#getBoolean(String)}
	*/
	public void setDisabled(String disabled)
	{
		this.disabled = Values.getBoolean(disabled);
	}
	
	/**
		Create the bean object. Call all setter methods that
		corresponds to parameters found in the XML file.
		@param actionClass The interface that the proxy must implement
	*/
	private synchronized void initBean(Class<? extends Action> actionClass)
	{
		if (initialised) return;
		
		try
		{
			// Create the bean
			bean = createBean(actionClass);
	
			// Set parameters on the bean
			actions = new Action[] { bean };
			for (Map.Entry<String, String> entry : parameters.entrySet())
			{
				String name = entry.getKey();
				String value = entry.getValue();
				
				// Find a setParam() method. Ignore parameter if not found
				Method setter = findSetterMethod(name, bean.getClass());						
				if (setter == null)
				{
					/* #### CONTINUE-STATEMENT #### */
					continue;
				}
				
				// Convert value to correct type for setParam(T) method
				Object o = convertValue(value, setter.getParameterTypes()[0]);
				try
				{
					setter.invoke(bean, o);
				}
				catch (IllegalAccessException ex)
				{}
				catch (InvocationTargetException ex)
				{}
			}
		}
		finally
		{
			initialised = true;
		}
	}

	/**
		Create the bean instance.
		@return A bean instance
	*/
	private Action createBean(Class<? extends Action> actionClass)
	{
		try
		{
			ClassLoader loader = actionClass.getClassLoader();		
			Class beanClass = ClassUtil.checkAndLoadClass(loader, beanClassName, true, actionClass);
			Action bean = actionClass.cast(beanClass.newInstance());
			return bean;
		}
		catch (Exception ex)
		{
			throw new IllegalArgumentException("Could not create bean: " + beanClassName, ex);
		}
	}
	
	/**
		Find a setter method for the specified parameter.
		Try: setParam(String), T getParam() or T isParam() --> setParam(T)
	*/
	private Method findSetterMethod(String parameterName, Class<?> beanClass)
	{
		Method setter = null;
		String setterMethodName = getMethodName("set", parameterName);
		
		// Look for setParam(String)
		setter = getMethod(beanClass, setterMethodName, String.class);
		if (setter != null) return setter;
		
		// Not found - Try: getParam()
		String getterMethodName = getMethodName("get", parameterName);
		Method getter = getMethod(beanClass, getterMethodName, null);
		
		if (getter == null)
		{
			// Last chance - Try: isParam()
			getterMethodName = getMethodName("is", parameterName);
			getter = getMethod(beanClass, getterMethodName, null);
		}
	
		if (getter != null)
		{
			setter = getMethod(beanClass, setterMethodName, getter.getReturnType());
		}
		
		return setter;
	}
	
	/**
		Convert the parameter name to a method name.
		First character of parameter name is upper-cased. The prefix
		is then prepended. Ex, 'icon' --> 'setIcon'
		@param parameterName The parameter name
		@return The tag name prefixed with 'set' and the first letter captialized
	*/
	private String getMethodName(String prefix, String parameterName)
	{
		String firstLetter = parameterName.substring(0, 1).toUpperCase();
		String setterName = prefix + firstLetter + parameterName.substring(1);
		return setterName;
	}	
	
	/**
		Get a method with a specific name with optionally, takes
		a single parameter of given type.
		@param beanClass The class to look for the method in
		@param name The name of the method
		@param param The type of the parameter, or null to look for
			a method with not parameters
		@return The Method or null if no such method can be found
	*/
	private Method getMethod(Class<?> beanClass, String name, Class<?> param)
	{
		Method method = null;
		try
		{
			Class[] params = param == null ? null : new Class[] { param };
			method = beanClass.getDeclaredMethod(name, params);
		}
		catch (NoSuchMethodException ex)
		{}
		return method;
	}
	
	/**
		Convert a string value to another type.
		@param sValue The value to convert
		@param type The class to convert the string value to
		@return The converted value, or null if it couldn't be converted
	*/
	private Object convertValue(String sValue, Class<?> type)
	{
		// The value to return
		Object value = null;

		// Convert to proper object type
		value = StringConverter.convertString(type, sValue);
		
		// Save the converted value in the cache
		return value;
	}
}
