/**
	$Id$

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager.processor;

import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jdom.Element;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemKey;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFileProcessor;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.PluginInfoKey;
import net.sf.basedb.util.extensions.manager.ProcessResults;
import net.sf.basedb.util.extensions.xml.PluginDefinitionFilter;
import net.sf.basedb.util.extensions.xml.PluginInfo;
import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	File processor implementation that load and install
	plug-ins.
	
	@author Nicklas
	@since 3.0
	@base.modified $Date$
*/
public class PluginInstallationProcessor
	implements ExtensionsFileProcessor
{

	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.util.extensions.manager.PluginInstallationProcessor");

	private final DbControl dc;
	private final XmlLoader loader;
	private final ProcessResults results;
	private int numFiles;
	private int numError;
	private int numTotalPlugins;

	/**
		Create a new processor. 
		
		@param dc Optional DbControl to use when installing plug-ins
			(if not given, only the information from the XML files are loaded,
			the plug-ins are not installed)
	*/
	public PluginInstallationProcessor(DbControl dc, ProcessResults results)
	{
		this.dc = dc;
		this.loader = new XmlLoader();
		this.results = results;
	}

	/*
		From the ExtensionsFileProcessor interface
		------------------------------------------
	*/
	@Override
	public void begin(ExtensionsManager manager, int numFiles) 
	{
		this.numFiles = 0;
		this.numError = 0;
		this.numTotalPlugins = 0;
		loader.setFilter(new PluginDefinitionFilter());
	}

	@SuppressWarnings("unchecked")
	@Override
	public void processFile(ExtensionsManager manager, WriteableExtensionsFile wFile) 
	{
		InputStream in = null;
		ExtensionsFile xtFile = wFile.getExtensionsFile();
		String jarFile = xtFile.isJar() ? xtFile.getName() : null;
		try
		{
			log.info("Loading plug-in definitions from file: " + xtFile);
			
			in = xtFile.getXmlStream();
			loader.loadXmlFile(in, xtFile.getName(), xtFile.getClassLoader(), true);
			FileUtil.close(in);
			List<PluginInfo> plugins = loader.getPluginDefinitions();
			
			int numInstalled = 0;
			int numUpdated = 0;
			int numProcessed = 0;
			int numConfigurations = 0;
			Map<String, PluginDefinition> allPlugins = new HashMap<String, PluginDefinition>();
			if (plugins.size() > 0)
			{
				PluginDefinition.checkInstallation(dc, plugins);
				
				ItemKey shareToEveryone = null;
				if (dc != null && dc.getSessionControl().hasSystemPermission(Permission.SHARE_TO_EVERYONE))
				{
					shareToEveryone = ItemKey.getShareToEveryOneKey(dc, Permission.USE);
				}
				
				wFile.open();
				
				for (PluginInfo info : plugins)
				{
					boolean deprecated = Values.getBoolean(info.getProperty("deprecated"));
					if (dc != null)
					{
						PluginDefinition plugin = PluginDefinition.installOrUpdate(dc, info, jarFile, shareToEveryone);
						if (plugin != null)
						{
							allPlugins.put(info.getClassName(), plugin);
							if (plugin.isInDatabase())
							{
								numUpdated++;
							}
							else
							{
								numInstalled++;
							}
						}
					}
					numProcessed++;
					if (!deprecated || !info.isNotFound())
					{
						wFile.registerObject(new PluginInfoKey(info), info);
					}
				}
			}
			
			// Load plug-in configurations
			if (dc != null)
			{
				in = xtFile.getStream("META-INF/plugin-configurations.xml");
				if (in != null)
				{
					List<Element> configurations = PluginConfiguration.loadXmlFile(in, "META-INF/plugin-configurations.xml");
					FileUtil.close(in);
					for (Element configuration : configurations)
					{
						String name = configuration.getChildText("configname");
						String description = configuration.getChildText("description");
						String pluginClassName = configuration.getAttributeValue("pluginClassName");
						
						PluginDefinition plugin = allPlugins.get(pluginClassName);
						if (plugin != null)
						{
							PluginConfiguration config = null;
							if (plugin.isInDatabase())
							{
								ItemQuery<PluginConfiguration> query = plugin.getPluginConfigurations();
								query.restrict(Restrictions.eq(Hql.property("name"), Expressions.parameter("name", name, Type.STRING)));
								List<PluginConfiguration> existing = query.list(dc);
								if (existing.size() == 0)
								{
									config = plugin.newPluginConfiguration();
								}
							}
							else
							{
								config = plugin.newPluginConfiguration();
							}
							if (config != null && !config.isInDatabase())
							{
								dc.saveItem(config);
								config.setItemKey(plugin.getItemKey());
								config.setProjectKey(plugin.getProjectKey());
								config.setName(name);
								config.setDescription(description);
								config.setParameterValues((List<Element>)configuration.getChildren("parameter"));
								numConfigurations++;
							}
						}
	
					}
				}
			}
			
			numFiles++;
			numTotalPlugins += numInstalled + numUpdated;
			if (results != null)
			{
				if (numInstalled > 0) results.addMessage(xtFile, numInstalled + " plug-in(s) installed.");
				if (numUpdated > 0) results.addMessage(xtFile, numUpdated + " plug-in(s) updated.");
				if (numConfigurations > 0) results.addMessage(xtFile, numConfigurations + " plug-in configuration(s) created");
				if (numInstalled == 0 && numUpdated == 0 && numProcessed > 0)
				{
					results.addMessage(xtFile, numProcessed + " plug-in(s) loaded");
				}
			}
			log.info("Installed/updated/processed " + numInstalled + "/" + numUpdated + "/" + numProcessed + " plug-ins from file: " + xtFile);
		}
		catch (Exception ex)
		{
			wFile.setError(true);
			numError++;
			if (results != null)
			{
				results.addErrorMessage(xtFile, "Failed to install plug-in(s): " + ex.getMessage());
			}
			log.error("Failed to install plug-ins from file: " + xtFile, ex);
		}
		finally
		{
			wFile.close();
			FileUtil.close(in);
		}

	}

	@Override
	public void done(ExtensionsManager manager) 
	{}

	@Override
	public void done(ExtensionsManager manager, Throwable t)
	{}
	// ------------------------------------------
	
	/**
		Get the loader the processor is using for parsing xml files with
		plug-in definitions.
	*/
	public XmlLoader getXmlLoader()
	{
		return loader;
	}
	
	public ProcessResults getProcessResults()
	{
		return results;
	}

	/**
		Get the number of files that was successfully processed.
	*/
	public int getNumFiles()
	{
		return numFiles;
	}
	
	/**
		Get the number of files that was had an error.
	*/
	public int getNumError()
	{
		return numError;
	}

	/**
		Get the total number of plug-in installed or updated
		by the processor.
	*/
	public int getNumPlugins()
	{
		return numTotalPlugins;
	}
	
}
