/**
	$Id: RawBioAssayValidator.java 5813 2011-10-18 09:24:06Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for raw bioassays. Validation rules:
	<ul>
	<li>Non-default raw data type: {@link Validator#NONDEFAULT_RAWDATATYPE}
	<li>Parent extract mismatch:  {@link Validator#PARENT_EXTRACT_MISMATCH}
	<li>The number of raw data spots is different from the number of
		features on the array design: {@link Validator#NONMATCHING_SPOTCOUNT}.
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-18 11:24:06 +0200 (Tue, 18 Oct 2011) $
*/
public class RawBioAssayValidator
	extends NameableNodeValidator<RawBioAssay>
{
	
	public RawBioAssayValidator()
	{
		super(null, null);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		
		// Check raw data type against default raw data type
		Project project = context.getProject();
		RawBioAssay rba = (RawBioAssay)node.getItem(dc);
		if (project != null)
		{
			project = Project.getById(dc, project.getId());
			RawDataType projectDefault = project.getDefaultRawDataType();
			if (projectDefault != null && rba.getRawDataType() != projectDefault)
			{
				context.createFailure(Validator.NONDEFAULT_RAWDATATYPE, node, null,
						new Fix("Change default raw data type of project", project)
					);
			}
		}
		
		// Check that parent items make sense
		DerivedBioAssay parentBioAssay = null;
		Extract parentBioAssayExtract = null;
		Extract parentExtract = null;
		PhysicalBioAssay physicalBioAssay = null;
		try
		{
			parentBioAssay = rba.getParentBioAssay();
			parentExtract = rba.getParentExtract();
			if (parentBioAssay != null) 
			{
				parentBioAssayExtract = parentBioAssay.getExtract();
				physicalBioAssay = parentBioAssay.getPhysicalBioAssay();
			}
		}
		catch (PermissionDeniedException ex)
		{}
		
		// Check that parent extract is same on the raw bioassay and derived bioassay
		if (parentBioAssayExtract != null)
		{
			if (!parentBioAssayExtract.equals(parentExtract))
			{
				context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
					"The parent bioassay '" + parentBioAssay.getName() + "' has different extract: " + 
						parentBioAssayExtract.getName(), 
					new Fix("Change extract on raw bioassay", rba),
					new Fix("Change extract on parent bioassay", parentBioAssay)
				);
			}
		}
		
		// Check that the extract is part of the physical bioassay
		if (parentExtract != null && physicalBioAssay != null)
		{
			if (!physicalBioAssay.getCreationEvent().isSource(parentExtract))
			{
				context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
					"The extract '" + parentExtract.getName() + "' is not found on the physical bioassay: " + 
						physicalBioAssay.getName(),
					new Fix("Change extract or parent bioassay on raw bioassay", rba),
					new Fix("Change extracts on physical bioassay", physicalBioAssay)
				);
			}
		}
				
		if (rba.hasArrayDesign() && rba.hasData())
		{
			try
			{
				ArrayDesign design = rba.getArrayDesign();
				String threshold = context.getValidationOptions().getOption(
						Validator.NONMATCHING_SPOTCOUNT, "threshold", "5%");
				
				int numDbSpots = rba.getNumDbSpots();
				int numDbFeatures = design.getNumDbFeatures();
				int numFileSpots = rba.getNumFileSpots();
				int numFileFeatures = design.getNumFileFeatures();
				int maxDbDelta = 0;
				int maxFileDelta = 0;
				if (threshold.endsWith("%"))
				{
					int percentage = Values.getInt(threshold.substring(0, threshold.length()-1));
					if (numDbFeatures > 0) maxDbDelta = (numDbFeatures * percentage) / 100;
					if (numFileFeatures > 0) maxFileDelta = (numFileFeatures * percentage) / 100;
				}
				else
				{
					maxDbDelta = Values.getInt(threshold);
					maxFileDelta = maxDbDelta;
				}
				
				
				if (numDbSpots > 0 && Math.abs(numDbFeatures - numDbSpots) > maxDbDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (db) <> features by more than " + threshold + ": " + 
						numDbSpots  + "<>" + numDbFeatures);
				}
				if (numFileSpots > 0 && Math.abs(numFileFeatures - numFileSpots) > maxFileDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (file) <> features by more than " + threshold + ": " + 
						numFileSpots  + "<>" + numFileFeatures);
				}
				
			}
			catch (RuntimeException ex)
			{}
		}
	}

	/**
		@return Always null
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		return null;
	}
	// ----------------------------
}
