/**
	$Id: ConnectionParameters.java 5618 2011-04-28 10:59:18Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.uri;

import java.net.URI;
import java.net.URISyntaxException;

import net.sf.basedb.core.FileServer;
import net.sf.basedb.core.data.FileServerData;
import net.sf.basedb.util.Values;

/**
	Connection parameters, such as username/password, that may be needed 
	to connect to a server.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-04-28 12:59:18 +0200 (Thu, 28 Apr 2011) $
*/
public class ConnectionParameters
{

	/**
		Create a parameters object using information from the
		{@link FileServer} object.
	*/
	public static ConnectionParameters create(FileServerData fileServer)
	{
		if (fileServer == null) return null;
		
		ConnectionParameters parameters = new ConnectionParameters();
		parameters.setConnectionManagerFactoryId(fileServer.getConnectionManagerFactory());
		parameters.setHost(fileServer.getHost());
		parameters.setUsername(fileServer.getUsername());
		parameters.setPassword(fileServer.getPassword());
		parameters.setServerCertificate(fileServer.getServerCertificate());
		parameters.setClientCertificate(fileServer.getClientCertificate());
		parameters.setClientCertificatePassword(fileServer.getClientCertificatePassword());
		return parameters;
	}
	
	private String factoryId;
	private String host;
	private String username;
	private String password;
	private byte[] serverCertificate;
	private byte[] clientCertificate;
	private String clientCertificatePassword;
	
	/**
		Create a new empty parameters object.
	*/
	public ConnectionParameters()
	{}
	
	/**
		Get the connection manager factory that we should use to
		access the file, or null to use auto-detection.
	*/
	public String getConnectionManagerFactoryId()
	{
		return factoryId;
	}
	public void setConnectionManagerFactoryId(String factoryId)
	{
		this.factoryId = factoryId;
	}
	
	/**
		Get the host:port that should override the host:port
		specified by the file URI.
	*/
	public String getHost()
	{
		return host;
	}
	public void setHost(String host)
	{
		this.host = host;
	}
	
	/**
		Get the username that should be used to authenticate the user.
	*/
	public String getUsername()
	{
		return username;
	}
	public void setUsername(String username)
	{
		this.username = username;
	}
	
	/**
		Get the password that should be used to authenticate the user.
	*/
	public String getPassword()
	{
		return password;
	}
	public void setPassword(String password)
	{
		this.password = password;
	}
	
	/**
		Get the public server certificate that indicates that we should only trust the 
		server if it can present a certificate that matches this. This property is typically
		only needed when using SSL to access servers that has a self-signed certificate. Servers
		which has a certificate signed by a trusted certification authority should be
		automatically trusted by BASE without this property. The certificate
		is a X.509 certificate in either binary or base64-encoded DER format.
		
		@return A byte array with the certificate or null if no certificate has been set
		@see FileServer#getServerCertificate()
	*/
	public byte[] getServerCertificate()
	{
		return serverCertificate;
	}
	public void setServerCertificate(byte[] certificate)
	{
		this.serverCertificate = certificate;
	}
	
	/**
		Get the client certificate that BASE should use to authenticate with the server
		when connecting using SSL. This property is typically needed when the server
		requires clients to authenticate themselves using a certificate. The certificate
		must be a PKCS #12 certificate in binary format, and it should contain only
		one certificate. 
		
		@return A byte array with the certificate or null if no certificate has been set
		@see FileServer#getClientCertificate()
	*/
	public byte[] getClientCertificate()
	{
		return clientCertificate;
	}
	public void setClientCertificate(byte[] certificate)
	{
		this.clientCertificate = certificate;
	}

	/**
		Get the password that is needed to unlock the client certificate. 
		@return The password (which may be null)
	*/
	public String getClientCertificatePassword()
	{
		return clientCertificatePassword;
	}
	public void setClientCertificatePassword(String password)
	{
		this.clientCertificatePassword = password;
	}

	/**
		If the connection parameters include a specific host:port a new URI
		is created were the host:port is replaced. Otherwise the original URI
		is returned unmodified.
		@param uri The URI to change
		@return The new URI
	*/
	public URI changeHost(URI uri)
	{
		String host = getHost();
		if (host == null) return uri;
		
		int port = -1;
		int colon = host.indexOf(':');
		if (colon > 0)
		{
			port = Values.getInt(host.substring(colon+1), -1);
			host = host.substring(0, colon);
		}
		try
		{
			uri = new URI(uri.getScheme(), uri.getUserInfo(), host, port, uri.getPath(), uri.getQuery(), uri.getFragment());
		}
		catch (URISyntaxException ex)
		{} // Shouldn't happen if the original URI is ok
		return uri;
	}
	
}
