/**
	$Id: ManualTransformCreator.java 5719 2011-09-06 11:56:29Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.AnnotationSetterPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.bfs.BaseInputStreamLocator;
import net.sf.basedb.util.bfs.GenericExtraFileImporter;
import net.sf.basedb.util.bfs.MetadataParser;
import net.sf.basedb.util.importer.BaseFileWrapper;
import net.sf.basedb.util.importer.spotdata.BaseFileImporter;
import net.sf.basedb.util.importer.spotdata.BfsImporter;

/**
	Plug-in that allows a user to manually register an external analysis
	procedure and to upload files that was generated by it.
	
	@author Nicklas
	@since 2.12
*/
public class ManualTransformCreator 
	extends AbstractAnalysisPlugin 
	implements InteractivePlugin, AnnotationSetterPlugin
{

	private RequestInformation configureJob;
	private RequestInformation configurePlugin;
	private List<PluginParameter<?>> toolParameters;
	
	private List<String> importableFileTypes = Arrays.asList(
				DataFileType.BASEFILE_MATRIX, 
				DataFileType.BASEFILE_SERIAL,
				DataFileType.BFS_METADATA);
	
	public ManualTransformCreator()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void run(Request request, Response response, ProgressReporter progress) 
	{
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			
			// Get parameters
			BioAssaySet source = getSourceBioAssaySet(dc);
			Job currentJob = getCurrentJob(dc);
			String transformationName = (String)job.getValue(TRANSFORMATION_NAME);
			String childName = (String)job.getValue(CHILD_NAME);
			String childDescription = (String)job.getValue(CHILD_DESCRIPTION);
			boolean copyAnnotations = Boolean.TRUE.equals(job.getValue(COPY_ANNOTATIONS));
			boolean keepSpotData = "keep".equals(job.getValue("spotData"));
			boolean importSpotData = "import".equals(job.getValue("spotData"));
			boolean useBase1ColumnNames = "base1".equals(job.getValue("columnNames"));
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			boolean allowMoreFiles = Boolean.TRUE.equals(configuration.getValue("allowMoreFiles"));
			
			// Create child items
			Transformation t = source.newTransformation(currentJob);
			t.setName(transformationName);
			dc.saveItem(t);
			BioAssaySet child = null;
			if (importSpotData)
			{
				String fileType = getImportFileType();
				if (fileType == null)
				{
					throw new ItemNotFoundException("You have not selected a file to import from");
				}
				File file = (File)job.getValue("file." + fileType);
				file = File.getById(dc, file.getId());
				if (fileType.startsWith("basefile."))
				{
					BaseFileImporter importer = new BaseFileImporter();
					if (!useBase1ColumnNames)
					{
						importer.useBase2ColumnNames(source.getRawDataType().getChannels());
					}
					importer.setDbControl(dc);
					importer.setSourceFileWrapper(new BaseFileWrapper(file));
					importer.setProgressReporter(progress);
					importer.setTransformation(t);
					child = importer.doImport();
				}
				else if (fileType.startsWith("bfs."))
				{
					Directory dir = file.getDirectory();
					BfsImporter importer = new BfsImporter();
					importer.setDbControl(dc);
					importer.setMetadataParser(MetadataParser.create(file));
					importer.setInputStreamLocator(new BaseInputStreamLocator(dc, dir));
					importer.setExtraFileImporter(new GenericExtraFileImporter(dc, dir, false));
					importer.setProgressReporter(progress);
					importer.setTransformation(t);
					child = importer.doImport();
				}
				else 
				{
					throw new InvalidDataException("Unknown file type: " + fileType);
				}
			}
			if (child == null)
			{
				child = t.newProduct(null, keepSpotData ? null : "new", keepSpotData);
				dc.saveItem(child);
			}
			child.setName(childName);
			child.setDescription(childDescription);
			
			// Copy annotations
			if (copyAnnotations)
			{
				PluginConfiguration cfg = getCurrentConfiguration(dc);
				if (cfg.isAnnotated()) child.getAnnotationSet().copyFrom(cfg, false);
			}
			
			// Set files
			if (fileTypes != null && fileTypes.size() > 0)
			{
				FileSet fs = child.getFileSet();
				for (DataFileType ft : fileTypes)
				{
					List<File> files = (List<File>)job.getValues("file." + ft.getExternalId());
					if (files != null && files.size() > 0)
					{
						for (File f : files)
						{
							fs.addMember(f, ft);
						}
					}
				}
				fs.validate(dc);
			}
			if (allowMoreFiles)
			{
				List<File> moreFiles = (List<File>)job.getValues("morefiles");
				if (moreFiles != null)
				{
					for (File f : moreFiles)
					{
						AnyToAny ata = AnyToAny.getNew(dc, child, f, f.getName(), false);
						dc.saveItem(ata);
					}
				}
			}
			
			dc.commit();
			response.setDone("Bioassay set " + childName + " created successfully");
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// ---------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void configure(GuiContext context, Request request, Response response) 
	{
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			RequestInformation ri = getConfigurePluginParameters(context);
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}
			
			storeValues(configuration, request, ri.getParameter("fileTypes"));
			storeValue(configuration, request, ri.getParameter("allowMoreFiles"));
			storeValues(configuration, request, ri.getParameter("parameters"));
			response.setDone("Configuration complete");
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			RequestInformation ri = getConfigureJobParameters(context);
			List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
			if (errors != null)
			{
				response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
				return;
			}
			
			storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
			storeValue(job, request, ri.getParameter("spotData"));
			storeValue(job, request, ri.getParameter("columnNames"));
			storeValue(job, request, ri.getParameter(TRANSFORMATION_NAME));
			storeValue(job, request, ri.getParameter(CHILD_NAME));
			storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));
			storeValue(job, request, ri.getParameter(COPY_ANNOTATIONS));
			
			// Files
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			if (fileTypes != null)
			{
				for (DataFileType ft : fileTypes)
				{
					storeValues(job, request, ri.getParameter("file." + ft.getExternalId()));
				}
			}
			storeValues(job, request, ri.getParameter("morefiles"));

			// Parameters
			List<PluginParameter<?>> tp = getToolParameters((List<String>)configuration.getValues("parameters"));
			if (tp != null)
			{
				for (PluginParameter<?> pp : tp)
				{
					storeValue(job, request, pp);
				}
			}
			
			BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
			PluginConfiguration config = configuration.getPluginConfiguration();
			response.setSuggestedJobName("Manual transform '" + config.getName() +
				"' on bioassay set '" + bioAssaySet.getName() + "'");

			boolean importSpotData = "import".equals(job.getValue("spotData"));
			if (importSpotData)
			{
				String fileType = getImportFileType();
				if (fileType == null)
				{
					response.setError("You have not selected a file to import from", null);
				}
				else
				{
					response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				}
			}
			else
			{
				response.setExecuteImmediately("Job configuration complete", Job.ExecutionTime.SHORTEST, false);
			}
		}
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException 
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		return requestInformation;
	}
	// ----------------------------------------------
	/*
		From the AnnotationSetterPlugin interface
		-------------------------------------------
	*/
	/**
		Return a query that finds annotation types for bioassay sets.
	*/
	public ItemQuery<AnnotationType> getAnnotationTypes()
	{
		return AnnotationType.getQuery(Item.BIOASSAYSET);
	}
	public String getAnnotationMessage()
	{
		return "The selected annotations will be copied to the child bioassay set that " +
				"is created by the manual transform.";
	}
	// ----------------------------------------------

	@SuppressWarnings("unchecked")
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob != null) return configureJob;
		DbControl dc = null;
		List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
		try
		{
			dc = sc.newDbControl();
			// Get the configuration options
			PluginConfiguration config = getCurrentConfiguration(dc);
			boolean allowMoreFiles = Boolean.TRUE.equals(configuration.getValue("allowMoreFiles"));
			List<DataFileType> fileTypes = (List<DataFileType>)configuration.getValues("fileTypes");
			boolean configIsAnnotated = config.isAnnotated();
			BioAssaySet source = getCurrentBioAssaySet(dc);
			
			// Check if any of the used file types can be imported
			boolean hasImportableFileType = false;
			boolean hasBaseFileType = false;
			if (fileTypes != null)
			{
				for (DataFileType dft : fileTypes)
				{
					String fileTypeId = dft.getExternalId();
					if (fileTypeId.startsWith("basefile.")) hasBaseFileType = true;
					if (importableFileTypes.contains(fileTypeId))
					{
						hasImportableFileType = true;
						break;
					}
				}
			}
			boolean hasDbData = source != null && source.getNumSpots() > 0;
			
			// Source bioassay set
			parameters.add(getSourceBioAssaySetParameter(null, null));
			
			if (hasDbData || hasImportableFileType)
			{
				List<String> options = new ArrayList<String>();
				if (hasImportableFileType) options.add("import");
				if (hasDbData) options.add("keep");
				options.add("none");
				parameters.add(new PluginParameter<String>(
					"spotData",
					"Spot data",
					"Select the procedure for creating in-databaes spot data for the " +
					"child bioassay set:\n\n" +
					"import = import spot data from the selected file(s) " +
					"(this option is only available if the tool generates a BASEfile or " +
					"BFS data)\n" +
					"keep = keep the spot data from the parent bioassay set\n" +
					"none = do not create any spot data in the database",
					new StringParameterType(255, null, false, 1, 0, 0, options)
				));
				if (hasBaseFileType)
				{
					parameters.add(new PluginParameter<String>(
						"columnNames",
						"Column names",
						"If you have selected Spot data=import, you may specify which " +
						"type of BASEfile you have:\n\n" +
						"base1 = The file uses BASE 1 column names\n" +
						"base2 = The file uses BASE 2 column names",
						new StringParameterType(255, "base2", false, 1, 0, 0, Arrays.asList("base1", "base2"))
					));
				}
			}
			// Names of transformation, child bioassay set etc.
			String transformationName = getTransformationName(dc);
			parameters.add(getTransformationNameParameter(null, null, transformationName));
			parameters.add(getChildNameParameter(null, null, "After " + transformationName));
			parameters.add(getChildDescriptionParameter(null, null, null));

			// Annotations section
			if (configIsAnnotated)
			{
				parameters.add(annotationSection);
				parameters.add(getCopyAnnotationsParmeter(null, 
					"If selected, annotations are copied from the configuration to the " +
					"child bioassay set that is created by the manual transform.", true));
			}
			
			// Parameters section
			List<PluginParameter<?>> toolParameters = 
				getToolParameters((List<String>)configuration.getValues("parameters"));
			if (toolParameters != null && toolParameters.size() > 0)
			{
				parameters.add(new PluginParameter<String>(
						"parameterSection", 
						"Parameters", 
						"Enter parameter values relevent to running the external tool.", 
						null));
				parameters.addAll(toolParameters);
			}
			
			parameters.add(new PluginParameter<String>(
				"fileSection", 
				"Result files", 
				"Select the files that was generated by the external tool.", 
				null));
			// reset that current file, otherwise all parameters get the same 
			// file as the default value
			sc.getCurrentContext(Item.FILE).setId(0);
			if (fileTypes != null)
			{
				for (DataFileType ft : fileTypes)
				{
					parameters.add(new PluginParameter<File>(
							"file." + ft.getExternalId(),
							ft.getName(),
							ft.getDescription(),
							new FileParameterType(null, false, 0)
					));
				}
			}
			
			if (allowMoreFiles)
			{
				parameters.add(new PluginParameter<File>(
					"morefiles",
					"More files...",
					"Select any additional files that was generated by the external tool " +
					"that are not present as pre-defined file types.",
					new FileParameterType(null, false, 0)
				));
			}
			
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
		configureJob = new RequestInformation(
			Request.COMMAND_CONFIGURE_JOB,
			"Files and parameters",
			"Enter information about the child bioassay set, the configuration and " +
			"parameters for the external tool and select/upload files that " +
			"was generated as a result",
			parameters
		);		
		return configureJob;
	}
	
	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin != null) return configurePlugin;

		List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();

			parameters.add(new PluginParameter<String>(
					"parameters",
					"Parameters",
					"Add parameters that are relevant for the external tool. All parameters " +
					"are assumed to be string parameters. Examples:\n" +
					"name\n" +
					"name<newline>helptext\n" +
					"name=option1|option2|option3\n",
					new StringParameterType(65536, null, false, 0, 0, 0)
				));
			
			ItemQuery<DataFileType> query = DataFileType.getQuery();
			query.restrict(Restrictions.eq(
					Hql.property("itemType"), 
					Expressions.integer(Item.BIOASSAYSET.getValue())
				));
			query.order(Orders.asc(Hql.property("name")));
			List<DataFileType> fileTypes = new ArrayList<DataFileType>(query.list(dc));
			parameters.add(new PluginParameter<DataFileType>(
				"fileTypes",
				"File types",
				"Select all registered file types that the external tool generates as " +
				"output.",
				new ItemParameterType<DataFileType>(DataFileType.class, null, false, 0, fileTypes)
			));
			
			parameters.add(new PluginParameter<Boolean>(
				"allowMoreFiles",
				"Allow more files",
				"Enabled this option if the tool generates files that has not been " +
				"registered as file types with BASE. This allows a user to select " +
				"any number of files and attach those to the resulting transform.",
				new BooleanParameterType(null, false)
			));

			configurePlugin = new RequestInformation(
				Request.COMMAND_CONFIGURE_JOB,
				"Files and parameters",
				"This page lets you select which types of output files the external tool " +
				"generates and what parameters that can be given to external tool.",
				parameters);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configurePlugin;
	}
	
	private List<PluginParameter<?>> getToolParameters(List<String> values)
	{
		if (values == null) return null;
		if (toolParameters != null) return toolParameters;
		toolParameters = new ArrayList<PluginParameter<?>>(values.size());
		
		for (String value : values)
		{
			// Split on first newline
			if (value == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			String[] temp = value.split("(\\n|\\r)+", 2);
			
			String name = temp[0];
			String description = temp.length > 1 ? temp[1] : null;
			int i = name.indexOf('=');
			List<String> options = null;
			if (i >= 1)
			{
				options = Arrays.asList(name.substring(i+1).split("\\|"));
				name = name.substring(0, i);
			}
			
			toolParameters.add(new PluginParameter<String>(
					"parameter." + name, 
					name, 
					description, 
					new StringParameterType(options != null ? 255 : 65536, null, false, 1, 0, 0 , options)
				));
		}
		return toolParameters;
	}
	
	private String getImportFileType()
	{
		for (String fileType : importableFileTypes)
		{
			if (job.getValue("file." + fileType) != null) return fileType;
		}
		return null;
	}
}
