/**
	$Id: MultiPropertyIdMethod.java 5646 2011-05-27 11:50:50Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;

/**
	Identification method implementation that can use a multiple properties
	(eg, name, id, externalId, etc.) for identifying items. Each property is
	tried in turn and a result is returned as soon as the query returns at
	least one result.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2011-05-27 13:50:50 +0200 (Fri, 27 May 2011) $
*/
public class MultiPropertyIdMethod
	implements IdMethod
{

	/**
		Use the name or external ID for finding items.
	*/
	public static final MultiPropertyIdMethod NAME_OR_EXTERNALID = 
		new MultiPropertyIdMethod("nameOrExternalId", "Name or external ID", 
				new String[] { "name", "externalId" }, false, Type.STRING);

	/**
		Use the name or barcode for finding items.
	*/
	public static final MultiPropertyIdMethod NAME_OR_BARCODE = 
		new MultiPropertyIdMethod("nameOrBarcode", "Name or barcode", 
				new String[] { "name", "barcode" }, false, Type.STRING);

	/**
		Use the name or system Id for finding items.
		@since 3.0
	*/
	public static final MultiPropertyIdMethod NAME_OR_SYSTEMID = 
		new MultiPropertyIdMethod("nameOrSystemId", "Name or system ID", 
				new String[] { "name", "systemId" }, false, Type.STRING);
		
	private final String method;
	private final String[] properties;
	private final String title;
	private final boolean unique;
	private final Type valueType;
	
	/**
		Create a new id method.

		@param method The method name, which should be unique among all methods
		@param title The title to use for display purposes
		@param properties The name of the properties to filter on when using this
			id method
		@param unique If this property is guaranteed to be unique or not
		@param valueType The type of values
	*/
	public MultiPropertyIdMethod(String method, String title, String[] properties, 
			boolean unique, Type valueType)
	{
		this.method = method;
		this.title = title;
		this.unique = unique;
		this.properties = properties;
		this.valueType = valueType;
	}
	
	/*
		From the IdMethod interface
		---------------------------
	*/
	@Override
	public String getMethod()
	{
		return method;
	}
	
	@Override
	public String getTitle()
	{
		return title;
	}
	
	@Override
	public String getColumnMappingParameterName()
	{
		return null;
	}

	@Override
	public boolean isTrulyUnique()
	{
		return unique;
	}

	/**
		Adds a restriction to the query:
		<code>property[0] = :identifier OR property[1] = :identifier ....</code>
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query)
	{
		// (property[0] = identifier AND 0 = numTries) OR property[1] = identifier AND 1 = numTries) ...
		
		Restriction[] r = new Restriction[properties.length];
		Expression identifier = Expressions.parameter("identifier");
		Expression numTries = Expressions.parameter("numTries");
		for (int i = 0; i < properties.length; ++i)
		{
			r[i] = Restrictions.and(
					Restrictions.eq(Hql.property(properties[i]), identifier),
					Restrictions.eq(Expressions.integer(i), numTries)
			);
		}
		query.restrictPermanent(Restrictions.or(r));
		return query;
	}

	/**
		Find all items with <code>property[0] = identifier</code>,
		or property[1] = identifier, ....
	*/
	@Override
	public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
	{
		Type idType = getValueType();
		query.setParameter("identifier", idType.parseString(identifier), idType);		
		int i = 0;
		List<I> result = Collections.emptyList();
		while (i < properties.length && result.isEmpty())
		{
			query.setParameter("numTries", i, Type.INT);
			result = query.list(dc);
			i++;
		}
		return result;
	}
	// ---------------------------
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return title;
	}
	// --------------------------

	public Type getValueType()
	{
		return valueType;
	}
	
}
