/**
	$Id: TestBioPlateEvent.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.BioPlateEvent;
import net.sf.basedb.core.BioPlateEventParticipant;
import net.sf.basedb.core.BioPlateEventType;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.BioWell.LockMode;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemProxy;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Trashcan;

import java.util.Date;
import java.util.List;
import java.util.Set;

/**
	 @author Nicklas
	 @since 2.17
	 @base.modified $Date: 2011-08-11 11:09:06 +0200 (Thu, 11 Aug 2011) $
 */
public class TestBioPlateEvent 
{
	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}
	
	static boolean test_all()
	{
		write("++Testing bioplate events");
		write_header();
		
		// Create bioplate types and bioplates
		int samplePlateTypeId = TestBioPlateType.test_create("Sample plate", Item.SAMPLE, LockMode.LOCKED_AFTER_MOVE);
		int extractPlateTypeId = TestBioPlateType.test_create("Extract plate", Item.EXTRACT, LockMode.LOCKED_AFTER_MOVE);
		int storagePlateTypeId = TestBioPlateType.test_create("Storage plate", null, LockMode.UNLOCKED);
		int plateGeometryId = TestPlateGeometry.test_create(2, 4, false);
		int samplePlateId = TestBioPlate.test_create("Samples #1", plateGeometryId, samplePlateTypeId, 0, false);
		int extractPlateId = TestBioPlate.test_create("Extracts #1", plateGeometryId, extractPlateTypeId, 0, false);
		int storagePlateId = TestBioPlate.test_create("Storage #1", plateGeometryId, storagePlateTypeId, 0, false);

		// Create biomaterial
		int sampleId1 = TestSample.test_create(0, "Foo.s1", false);
		int sampleId2 = TestSample.test_create(0, "Bar.s2", false);
		int sampleId3 = TestSample.test_create(0, "Ref.s3", false);

		int protocolId = TestProtocol.test_create(SystemItems.getId(Protocol.EXTRACTION), "Extract on plate", false);
		int hardwareId = TestHardware.test_create(SystemItems.getId(Hardware.OTHER), "Extraction bot", false);
		
		// Place the biomaterial on a plate using an event
		int id1 = test_place_on_plate_event("Place samples", samplePlateId, Item.SAMPLE, sampleId1, sampleId2, sampleId3);
		test_load(id1);
		test_list_participants(id1, 1);
		
		// Create extracts from the samples on a reaction plate
		int id2 = test_extraction_event("Create extracts", samplePlateId, extractPlateId, protocolId, hardwareId);
		test_list_participants(id2, 2);
		
		// Move the samples to a storage plate
		int id3 = test_move_event("Move samples to storage", samplePlateId, storagePlateId);
		test_list_participants(id3, 2);
		
		test_list(-1);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		test_delete(id1);
		test_delete(id2);
		test_delete(id3);
		
		// Delete the plates and the biomaterial on them
		TestBioPlate.test_delete_contents(extractPlateId);
		TestBioPlate.test_delete_contents(storagePlateId);
		TestBioPlate.test_delete_contents(samplePlateId);

		TestHardware.test_delete(hardwareId);
		TestProtocol.test_delete(protocolId);

		TestBioPlate.test_delete(storagePlateId);
		TestBioPlate.test_delete(samplePlateId);
		TestBioPlate.test_delete(extractPlateId);
		TestPlateGeometry.test_delete(plateGeometryId);
		TestBioPlateType.test_delete(samplePlateTypeId);
		TestBioPlateType.test_delete(extractPlateTypeId);
		TestBioPlateType.test_delete(storagePlateTypeId);
		
		write("++Testing bioplate events "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();		
			BioPlateEvent event = BioPlateEvent.getById(dc, id);
			dc.deleteItem(event);
			Set<ItemProxy> using = event.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(event + " is used by " + using);
			}
			dc.commit();
			write("--Delete bioplate event OK");
		}
		catch (Throwable ex)
		{
			write("--Delete bioplate event FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<BioPlateEvent> events = BioPlateEvent.getQuery().list(dc);
			for (int i = 0; i<events.size(); i++)
			{
				write_item(i, events.get(i));
			}
			if (expectedResults >=0 && expectedResults != events.size())
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + events.size());
			}
			write ("--List bioplate events OK (" + events.size() +")");
		}
		catch (Throwable ex)
		{
			write("--List bioplate events FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlateEvent event = BioPlateEvent.getById(dc, id);
			write_item(0, event);
			write("--Load bioplate event OK");
		}
		catch (Throwable ex)
		{
			write("--Load bioplate event FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<BioPlateEvent> l = BioPlateEvent.getQuery().list(dc);
			int numDeleted = Trashcan.delete(dc.getSessionControl(), l, true, null);
			if (numDeleted != l.size()) throw new BaseException("Could not delete all items: " + numDeleted + " of " + l.size());
			dc.commit();
			write("--Delete all bioplate events OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all bioplate events FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			System.out.println("   \tID \tName      \tEvent type\tEvent date\tEntry date\tDescription\tProtocol\tHardware");
			System.out.println("-- \t-- \t----------\t----------\t----------\t----------\t-----------\t--------\t--------");
		}
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void write_item(int i, BioPlateEvent event)
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+event.getId()+"\t"+event.getName()+"\t"+event.getEventType()+
				"\t" + event.getEventDate()+"\t"+event.getEntryDate()+"\t"+event.getDescription()+
				"\t"+event.getProtocol()+"\t"+event.getHardware());
	}
	
	static void write_item(int i, BioPlateEventParticipant participant)
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+participant.getId()+"\t"+participant.getRole()+
				"\t"+participant.getIndex()+"\t" + participant.getBioPlate());
	}

	
	static int test_place_on_plate_event(String name, int bioPlateId, Item bioMaterialType, int... bioMaterialIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.BIOPLATEEVENT) || bioPlateId == 0)
		{
			return 0;
		}
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			
			BioPlate plate = BioPlate.getById(dc, bioPlateId);
			BioPlateEventType eventType = BioPlateEventType.getById(dc, BioPlateEventType.PLACE_ON_PLATE);
			
			BioPlateEvent event = BioPlateEvent.getNew(dc, eventType);
			event.setName(name);
			event.setEventDate(new Date());
			BioPlateEventParticipant destination = event.addParticipant(plate, "destination", 1);
			dc.saveItem(event);
			
			if (bioMaterialType != null && bioMaterialIds != null)
			{
				List<BioWell> wells = plate.getBioWells().list(dc);
				for (int i = 0; i < wells.size() && i < bioMaterialIds.length; ++i)
				{
					MeasuredBioMaterial bm = (MeasuredBioMaterial)bioMaterialType.getById(dc, bioMaterialIds[i]);
					BioWell well = wells.get(i);
					bm.setBioWell(well);
					BioMaterialEvent bmEvent = bm.newEvent(destination);
					bmEvent.setComment("Well: " + well.getCoordinate());
					dc.saveItem(bmEvent);
				}
			}
			dc.commit();
			id = event.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(event, false);
			write_item(0, event);
			write("--Create 'place-on-plate' event OK");
		}
		catch (Throwable ex)
		{
			write("--Create 'place-on-plate' event FAIILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_extraction_event(String name, int samplePlateId, int extractPlateId, int protocolId, int hardwareId)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.BIOPLATEEVENT) || samplePlateId == 0 || extractPlateId == 0)
		{
			return 0;
		}
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			
			BioPlate samples = BioPlate.getById(dc, samplePlateId);
			BioPlate extracts = BioPlate.getById(dc, extractPlateId);
			BioPlateEventType eventType = BioPlateEventType.getById(dc, BioPlateEventType.CREATE_BIOMATERIAL);
			
			// Create the master event
			BioPlateEvent event = BioPlateEvent.getNew(dc, eventType);
			event.setName(name);
			event.setEventDate(new Date());
			if (protocolId != 0)
			{
				event.setProtocol(Protocol.getById(dc, protocolId));
			}
			if (hardwareId != 0)
			{
				event.setHardware(Hardware.getById(dc, hardwareId));
			}
			// Add participating plates
			BioPlateEventParticipant source = event.addParticipant(samples, "source", 1);
			BioPlateEventParticipant destination = event.addParticipant(extracts, "destination", 1);
			dc.saveItem(event);
			
			// Create child extracts
			for (BioWell well : samples.getBioWells().list(dc))
			{
				Sample s = (Sample)well.getBioMaterial();
				if (s != null)
				{
					// Create a new extract and put it in the same coordinate on the
					// extracts plate
					Extract e = s.newExtract(null, destination);
					e.setBioWell(extracts.getBioWell(well.getPlateCoordinate()));
					e.setName(s.getName()+".e"+s.countChildren(false));
					e.getCreationEvent().setComment("Well: " + well.getCoordinate());
					dc.saveItem(e);
				}
			}
			
			dc.commit();
			id = event.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(event, false);
			write_item(0, event);
			write("--Create extraction plate event OK");
		}
		catch (Throwable ex)
		{
			write("--Create extraction plate event FAIILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_move_event(String name, int sourcePlateId, int destinationPlateId)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.BIOPLATEEVENT) || sourcePlateId == 0 || destinationPlateId == 0)
		{
			return 0;
		}
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			
			BioPlate source = BioPlate.getById(dc, sourcePlateId);
			BioPlate destination = BioPlate.getById(dc, destinationPlateId);
			BioPlateEventType eventType = BioPlateEventType.getById(dc, BioPlateEventType.MOVE);
			
			// Create the master event
			BioPlateEvent event = BioPlateEvent.getNew(dc, eventType);
			event.setName(name);
			event.setEventDate(new Date());

			// Add participating plates
			event.addParticipant(source, "source", 1);
			BioPlateEventParticipant destParticipant = event.addParticipant(destination, "destination", 1);
			dc.saveItem(event);
			
			// Move the biomaterial
			for (BioWell well : source.getBioWells().list(dc))
			{
				MeasuredBioMaterial bm = well.getBioMaterial();
				if (bm != null)
				{
					// Move the sample to the storage plate
					bm.setBioWell(destination.getBioWell(well.getRow(), well.getColumn()));
					BioMaterialEvent bmEvent = bm.newEvent(destParticipant);
					bmEvent.setComment(source.getName() + " [" + well.getCoordinate() + "] → "+
							destination.getName() + " [" + well.getCoordinate() + "]");
					dc.saveItem(bmEvent);
				}
			}
			
			dc.commit();
			id = event.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(event, false);
			write_item(0, event);
			write("--Create move event OK");
		}
		catch (Throwable ex)
		{
			write("--Create move event FAIILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	
	static void test_list_participants(int eventId, int expectedResults)
	{
		if (eventId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlateEvent event = BioPlateEvent.getById(dc, eventId);
			ItemResultList<BioPlateEventParticipant> participants = event.getParticipants().list(dc);
			for (int i = 0; i<participants.size(); i++)
			{
				write_item(i, participants.get(i));
			}
			if (expectedResults >=0 && expectedResults != participants.size())
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + participants.size());
			}
			write ("--List bioplate event participants OK (" + participants.size() +")");
		}
		catch (Throwable ex)
		{
			write("--List bioplate event participants FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}