import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemProxy;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Quantity;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Unit;
import net.sf.basedb.util.units.UnitUtil;

/*
 	$Id: TestUnit.java 5340 2010-05-10 06:15:00Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

public class TestUnit
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing units");
		
		test_split_value("123 mm", "123", "mm");
		test_split_value("123.6 mm", "123.6", "mm");
		test_split_value("123.cm", "123", "cm");
		test_split_value("-123mm", "-123", "mm");
		test_split_value("-0.44", "-0.44", null);
		test_split_value("-0.33cm", "-0.33", "cm");
		test_split_value("1.2345E10cm", "1.2345E10", "cm");
		test_split_value("1.2345E-10 mm", "1.2345E-10", "mm");
		test_split_value("1.2345e+10", "1.2345e+10", null);
		test_split_value("1.2345eV", "1.2345", "eV");
		
		write_header();
		// Standard tests: create, load, list
		int lengthId = SystemItems.getId(Quantity.LENGTH);
		int id1 = test_create(lengthId, "Astronimical unit", "ua", 1.49598E11, 0.0, true);
		test_load(id1);
		test_add_symbol(id1, "ae");
		test_load_by_symbol(lengthId, "ae");
		test_list();
		
		// Extra test, load by symbol, convert between units
		int meterId = test_load_by_symbol(lengthId, "m");
		test_convert_units(meterId, 55, id1, 3.67651973957E-10, 1E-6);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();

		// Standard test: Delete
		test_remove_symbol(id1, "ae");
		test_delete(id1);
		write("++Testing units "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}
	
	static int test_create(int quantityId, String name, String symbol, double factor, double offset, boolean setAll)
	{
		if (quantityId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.UNIT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Quantity q = Quantity.getById(dc, quantityId);
			Unit u = Unit.getNew(dc, q, symbol);
			if(setAll)
			{
				u.setName("Test unit");
				u.setDescription("Added at "+new Date());
			}
			if (name != null) u.setName(name);
			u.changeReferenceFactorAndOffset(factor, offset);
			dc.saveItem(u);
			dc.commit();
			id = u.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(u, false);
			write_item(0, u);
			write("--Create unit OK");
		}
		catch (Throwable ex)
		{
			write("--Create unit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Unit u = Unit.getById(dc, id);
			dc.deleteItem(u);
			Set<ItemProxy> using = u.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(u + " is used by " + using);
			}
			dc.commit();
			write("--Delete unit OK");
		}
		catch (Throwable ex)
		{
			write("--Delete unit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Unit u = Unit.getById(dc, id);
			write_item(0, u);
			write("--Load unit OK");
		}
		catch (Throwable ex)
		{
			write("--Load unit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static int test_load_by_symbol(int quantityId, String symbol)
	{
		if (quantityId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Quantity q = Quantity.getById(dc, quantityId);
			Unit u = q.getUnit(symbol);
			id = u.getId();
			write_item(0, u);
			write("--Load unit by symbol OK");
		}
		catch (Throwable ex)
		{
			write("--Load unit by symbol FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_list()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Unit> l = Unit.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			write("--List units OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List units FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tQuantity\tSymbol\tFactor\tOffset\tAliases");
			write("-- \t-- \t--------- \t-----------\t--------\t------\t------\t------\t-------");
		}
	}
	static void write_item(int i, Unit u)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			write(i+":\t"+u.getId()+"\t"+u.getName()+"\t"+u.getDescription()+"\t"+
					u.getQuantity().getName() + "\t" +
					u.getDisplaySymbol()+"\t"+u.getReferenceFactor() + "\t" + 
					u.getReferenceOffset() + "\t" + u.getSymbols());
		}
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_convert_units(int unit1Id, double v1, int unit2Id, double v2, double delta)
	{
		if (unit1Id == 0 || unit2Id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Unit u1 = Unit.getById(dc, unit1Id);
			Unit u2 = Unit.getById(dc, unit2Id);
			
			double v1Tov2 = u1.toUnit(v1, u2);
			double v2Tov1 = u1.fromOther(v2, u2);
				
			if (Math.abs(v2 - v1Tov2) > delta)
			{
				throw new BaseException("Conversion of " + 
						v1 + u1.getDisplaySymbol() + " --> " + u2.getDisplaySymbol() + " failed. " +
						"Got " + v1Tov2 +"; expected " + v2);
			}
			if (Math.abs(v1 - v2Tov1) > delta)
			{
				throw new BaseException("Conversion of " + 
						v2 + u2.getDisplaySymbol() + " --> " + u1.getDisplaySymbol() + " failed. " +
						"Got " + v2Tov1 +"; expected " + v1);
			}
			if (!TestUtil.getSilent())
			{
				write("--" + v1 + u1.getDisplaySymbol() + " ~= " + v1Tov2 + u2.getDisplaySymbol());
				write("--" + v2 + u2.getDisplaySymbol() + " ~= " + v2Tov1 + u1.getDisplaySymbol());
			}
			
			write("--Convert units OK");
		}
		catch (Throwable ex)
		{
			write("--Convert units FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_add_symbol(int id, String symbol)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Unit u = Unit.getById(dc, id);
			u.addSymbol(symbol);
			dc.commit();
			dc = TestUtil.getDbControl();
			u = Unit.getById(dc, id);
			write_item(0, u);
			if (!u.getSymbols().contains(symbol))
			{
				throw new BaseException("Add symbol '" + symbol + "' to unit '" + u.getName() + "' failed.");
			}
			write("--Add symbol to unit OK");
		}
		catch (Throwable ex)
		{
			write("--Add symbol to unit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_remove_symbol(int id, String symbol)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Unit u = Unit.getById(dc, id);
			u.removeSymbol(symbol);
			dc.commit();
			dc = TestUtil.getDbControl();
			u = Unit.getById(dc, id);
			write_item(0, u);
			if (u.getSymbols().contains(symbol))
			{
				throw new BaseException("Remove symbol '" + symbol + "' from unit '" + u.getName() + "' failed.");
			}
			write("--Remove symbol from unit OK");
		}
		catch (Throwable ex)
		{
			write("--Remove symbol from unit FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_split_value(String value, String number, String unit)
	{
		String[] valueWithUnit = UnitUtil.splitValueWithUnit(value);
		boolean splitOk = false;
		if (number == null)
		{
			splitOk = valueWithUnit == null;
		}
		else if (valueWithUnit != null)
		{
			splitOk = number.equals(valueWithUnit[0]) && StringUtil.isEqualOrNull(unit, valueWithUnit[1]);
		}
		
		if (splitOk)
		{
			write("--" + value + " --> " + number + "[" + unit + "] OK");
		}
		else
		{
			String msg = "--" + value + " --> " + number + "[" + unit + "] FAILED: ";
			if (valueWithUnit == null) 
			{
				msg += "null";
			}
			else
			{
				msg += valueWithUnit[0] + "[" + valueWithUnit[1] + "]";
			}
			write(msg);
			ok = false;
		}
	}
}
