/**
	$Id: PluginUtil.java 5765 2011-09-27 13:38:36Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginExecutionRequest;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ConsoleProgressReporter;

/**

 @author nicklas
 @version 2.2
 @base.modified $Date: 2011-09-27 15:38:36 +0200 (Tue, 27 Sep 2011) $
 */
public class PluginUtil
{

	/**
		Create a new or load the existing plug-in configuration with a given name. Note!
		If the configuration already exists, the parameters are not updated.
	*/
	public static PluginConfiguration createPluginConfiguration(DbControl dc, String name, String className, Map<String, Object> parameters)
	{
		TestUtil.write("--Creating/loading plugin configuration: " + name + "\n");
		
		PluginDefinition definition = PluginDefinition.getByClassName(dc, className);
		ItemQuery<PluginConfiguration> query = definition.getPluginConfigurations();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		List<PluginConfiguration> result = query.list(dc);
		if (result.size() > 0) return result.get(0);
		
		PluginConfiguration config = PluginConfiguration.getNew(dc, definition);
		config.setName(name);
		if (parameters != null)
		{
			ParameterType stringType = new StringParameterType();
			ParameterType intType = new IntegerParameterType();
			ParameterType boolType = new BooleanParameterType();
			for (Map.Entry<String, Object> param : parameters.entrySet())
			{
				Object value = param.getValue();
				List<?> values = Collections.singletonList(value);
				if (value instanceof String)
				{
					config.setParameterValues(param.getKey(), stringType, values);
				}
				else if (value instanceof Integer)
				{
					config.setParameterValues(param.getKey(), intType, values);
				}
				else if (value instanceof Boolean)
				{
					config.setParameterValues(param.getKey(), boolType, values);
				}
				else
				{
					throw new BaseException("Invalid paramater type: " + value);
				}
			}
		}
		
		dc.saveItem(config);
		return config;
	}

	/**
		Import reporters.
	 	@param config The configuration to use on the plugin 
	 	@param file File item to import reporters from.
		@param jobName The name of the job
	 	@throws Exception If something fails.
	*/
	public static void importReporters(PluginConfiguration config, File file, String jobName)
		throws Exception
	{
		TestUtil.write("--Importing reporters from file: " + file.getName() + "\n");
		
		DbControl dc = TestUtil.getDbControl();
		try
		{
			config = PluginConfiguration.getById(dc, config.getId());
			Job job = Job.getNew(dc, config.getPluginDefinition(), config, null);
			job.setName(jobName);
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("updateExisting", new BooleanParameterType(), true);
			job.setParameterValue("stringTooLongError", new StringParameterType(), "crop");
			job.setParameterValue("invalidUseOfNullError", new StringParameterType(), "skip");
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	/**
		Execute a job.
		@param job
	*/
	public static void executeJob(Job job)
	{
		TestUtil.write("--Executing job: " + job.getName() + "\n");
		DbControl dc = TestUtil.getDbControl();
		job = Job.getById(dc, job.getId());
		PluginExecutionRequest request = 
			job.execute(new ConsoleProgressReporter(false), TestUtil.getLocalIp());
		dc.commit();
	
		PluginResponse response = request.invoke();
		if (response.getStatus() == Response.Status.ERROR)
		{
			throw new BaseException(response.getMessage(), response.getErrorList().get(0));
		}
		TestUtil.write("--Job ended successfully: " + response.getMessage() + "\n");
	}

}
